def expand_leading_tab(text: str, width: int = 4) -> str: ...
def expand_tab(text: str, space: str = "    ") -> str: ...
def escape(s: str, quote: bool = True) -> str: ...
def escape_url(link: str) -> str: ...
def safe_entity(s: str) -> str: ...
def unikey(s: str) -> str: ...
def unescape(s: str) -> str: ...
def striptags(s: str) -> str: ...
def strip_end(src: str) -> str: ...



import json
from typing import Any

import pytest

import schema_salad.metaschema as cg_metaschema
from schema_salad.exceptions import ValidationException
from schema_salad.utils import yaml_no_ts

from .matcher import JsonDiffMatcher
from .util import get_data_uri, get_path


def test_load() -> None:
    doc = {
        "type": "record",
        "fields": [{"name": "hello", "doc": "Hello test case", "type": "string"}],
    }
    rs = cg_metaschema.RecordSchema.fromDoc(
        doc, "http://example.com/", cg_metaschema.LoadingOptions(no_link_check=True)
    )
    assert "record" == rs.type_
    assert rs.fields and "http://example.com/#hello" == rs.fields[0].name
    assert "Hello test case" == rs.fields[0].doc
    assert "string" == rs.fields[0].type_
    assert {
        "type": "record",
        "fields": [
            {
                "name": "http://example.com/#hello",
                "doc": "Hello test case",
                "type": "string",
            }
        ],
    } == rs.save()


def test_err() -> None:
    doc = {"doc": "Hello test case", "type": "string"}
    with pytest.raises(ValidationException):
        cg_metaschema.RecordField.fromDoc(doc, "", cg_metaschema.LoadingOptions())


def test_include() -> None:
    doc = {"name": "hello", "doc": [{"$include": "hello.txt"}], "type": "documentation"}
    path_uri = get_data_uri("tests") + "/_"
    rf = cg_metaschema.Documentation.fromDoc(
        doc,
        "http://example.com/",
        cg_metaschema.LoadingOptions(fileuri=path_uri, no_link_check=True),
    )
    assert "http://example.com/#hello" == rf.name
    assert ["hello world!\n"] == rf.doc
    assert "documentation" == rf.type_
    assert {
        "name": "http://example.com/#hello",
        "doc": ["hello world!\n"],
        "type": "documentation",
    } == rf.save()


def test_import() -> None:
    doc = {"type": "record", "fields": [{"$import": "hellofield.yml"}]}
    lead = get_data_uri("tests")
    rs = cg_metaschema.RecordSchema.fromDoc(
        doc, "http://example.com/", cg_metaschema.LoadingOptions(fileuri=lead + "/_")
    )
    assert "record" == rs.type_
    assert rs.fields and lead + "/hellofield.yml#hello" == rs.fields[0].name
    assert "hello world!\n" == rs.fields[0].doc
    assert "string" == rs.fields[0].type_
    assert {
        "type": "record",
        "fields": [
            {
                "name": lead + "/hellofield.yml#hello",
                "doc": "hello world!\n",
                "type": "string",
            }
        ],
    } == rs.save()


def test_import2() -> None:
    path_uri = get_data_uri("tests/docimp/d1.yml")
    rs = cg_metaschema.load_document(path_uri, "", cg_metaschema.LoadingOptions())
    path2_uri = get_data_uri("tests/docimp/d1.yml")
    assert [
        {
            "doc": [
                "*Hello*",
                "hello 2",
                "*dee dee dee five*",
                "hello 3",
                "hello 4",
                "*dee dee dee five*",
                "hello 5",
            ],
            "type": "documentation",
            "name": path2_uri + "#Semantic_Annotations_for_Linked_Avro_Data",
        }
    ] == [r.save() for r in rs]


def test_err2() -> None:
    doc = {
        "type": "rucord",
        "fields": [{"name": "hello", "doc": "Hello test case", "type": "string"}],
    }
    with pytest.raises(ValidationException):
        cg_metaschema.RecordSchema.fromDoc(doc, "", cg_metaschema.LoadingOptions())


def test_idmap() -> None:
    doc = {
        "type": "record",
        "fields": {"hello": {"doc": "Hello test case", "type": "string"}},
    }
    rs = cg_metaschema.RecordSchema.fromDoc(
        doc, "http://example.com/", cg_metaschema.LoadingOptions(no_link_check=True)
    )
    assert "record" == rs.type_
    assert rs.fields and "http://example.com/#hello" == rs.fields[0].name
    assert "Hello test case" == rs.fields[0].doc
    assert "string" == rs.fields[0].type_
    assert {
        "type": "record",
        "fields": [
            {
                "name": "http://example.com/#hello",
                "doc": "Hello test case",
                "type": "string",
            }
        ],
    } == rs.save()


def test_idmap2() -> None:
    doc = {"type": "record", "fields": {"hello": "string"}}
    rs = cg_metaschema.RecordSchema.fromDoc(
        doc, "http://example.com/", cg_metaschema.LoadingOptions(no_link_check=True)
    )
    assert "record" == rs.type_
    assert rs.fields and "http://example.com/#hello" == rs.fields[0].name
    assert rs.fields[0].doc is None
    assert "string" == rs.fields[0].type_
    assert {
        "type": "record",
        "fields": [{"name": "http://example.com/#hello", "type": "string"}],
    } == rs.save()


def test_load_pt() -> None:
    path_uri = get_data_uri("tests/pt.yml")
    doc = cg_metaschema.load_document(
        path_uri, "", cg_metaschema.LoadingOptions(no_link_check=True)
    )
    assert [
        "https://w3id.org/cwl/salad#null",
        "http://www.w3.org/2001/XMLSchema#boolean",
        "http://www.w3.org/2001/XMLSchema#int",
        "http://www.w3.org/2001/XMLSchema#long",
        "http://www.w3.org/2001/XMLSchema#float",
        "http://www.w3.org/2001/XMLSchema#double",
        "http://www.w3.org/2001/XMLSchema#string",
    ] == doc.symbols


def test_shortname() -> None:
    """Test shortname() function."""
    assert cg_metaschema.shortname("http://example.com/foo") == "foo"
    assert cg_metaschema.shortname("http://example.com/#bar") == "bar"
    assert cg_metaschema.shortname("http://example.com/foo/bar") == "bar"
    assert cg_metaschema.shortname("http://example.com/foo#bar") == "bar"
    assert cg_metaschema.shortname("http://example.com/#foo/bar") == "bar"
    assert cg_metaschema.shortname("http://example.com/foo#bar/baz") == "baz"


@pytest.fixture
def metaschema_pre() -> Any:
    """Prep-parsed schema for testing."""
    with get_path("tests/metaschema-pre.yml").open() as f:
        pre = json.load(f)
    return pre


def test_load_metaschema(metaschema_pre: Any) -> None:
    path_uri = get_data_uri("metaschema/metaschema.yml")
    doc = cg_metaschema.load_document(
        path_uri,
        "",
        cg_metaschema.LoadingOptions(no_link_check=True),
    )
    saved = [d.save(relative_uris=False) for d in doc]
    # with open(get_data("tests/metaschema-pre.yml"), "w") as fh
    #    json.dump(saved, fh, indent=4)
    assert saved == JsonDiffMatcher(metaschema_pre)


def test_load_by_yaml_metaschema(metaschema_pre: Any) -> None:
    path = get_path("metaschema/metaschema.yml")
    with path.open() as path_handle:
        yaml = yaml_no_ts()
        yaml_doc = yaml.load(path_handle)
    doc = cg_metaschema.load_document_by_yaml(
        yaml_doc,
        path.as_uri(),
        None,
    )
    saved = [d.save(relative_uris=False) for d in doc]
    assert saved == JsonDiffMatcher(metaschema_pre)


def test_load_cwlschema() -> None:
    path_uri = get_data_uri("tests/test_schema/CommonWorkflowLanguage.yml")
    doc = cg_metaschema.load_document(
        path_uri,
        "",
        cg_metaschema.LoadingOptions(no_link_check=True),
    )
    path2 = get_path("tests/cwl-pre.yml")
    saved = [d.save(relative_uris=False) for d in doc]
    # with path2.open("w") as f:
    #     json.dump(saved, f, indent=2)
    with path2.open() as f:
        pre = json.load(f)
    assert saved == JsonDiffMatcher(pre)



from typing import (
    Any,
    ClassVar,
    Dict,
    Iterable,
    List,
    Match,
    MutableMapping,
    Optional,
    Pattern,
)

from .core import InlineState as InlineState
from .core import Parser as Parser
from .helpers import HTML_ATTRIBUTES as HTML_ATTRIBUTES
from .helpers import HTML_TAGNAME as HTML_TAGNAME
from .helpers import PREVENT_BACKSLASH as PREVENT_BACKSLASH
from .helpers import PUNCTUATION as PUNCTUATION
from .helpers import parse_link as parse_link
from .helpers import parse_link_label as parse_link_label
from .helpers import parse_link_text as parse_link_text
from .helpers import unescape_char as unescape_char
from .util import escape as escape
from .util import escape_url as escape_url
from .util import unikey as unikey

PAREN_END_RE: Pattern[str]
AUTO_EMAIL: str
INLINE_HTML: str
EMPHASIS_END_RE: Dict[str, Pattern[str]]

class InlineParser(Parser[InlineState]):
    sc_flag: int
    state_cls = InlineState
    STD_LINEBREAK: str
    HARD_LINEBREAK: str
    SPECIFICATION: ClassVar[Dict[str, str]]
    DEFAULT_RULES: ClassVar[Iterable[str]]
    hard_wrap: bool
    def __init__(self, hard_wrap: bool = False) -> None: ...
    def parse_escape(self, m: Match[str], state: InlineState) -> int: ...
    def parse_link(self, m: Match[str], state: InlineState) -> Optional[int]: ...
    def parse_auto_link(self, m: Match[str], state: InlineState) -> int: ...
    def parse_auto_email(self, m: Match[str], state: InlineState) -> int: ...
    def parse_emphasis(self, m: Match[str], state: InlineState) -> int: ...
    def parse_codespan(self, m: Match[str], state: InlineState) -> int: ...
    def parse_linebreak(self, m: Match[str], state: InlineState) -> int: ...
    def parse_softbreak(self, m: Match[str], state: InlineState) -> int: ...
    def parse_inline_html(self, m: Match[str], state: InlineState) -> int: ...
    def process_text(self, text: str, state: InlineState) -> None: ...
    def parse(self, state: InlineState) -> List[Dict[str, Any]]: ...
    def precedence_scan(
        self, m: Match[str], state: InlineState, end_pos: int, rules: Optional[List[str]] = None
    ) -> Optional[int]: ...
    def render(self, state: InlineState) -> List[Dict[str, Any]]: ...
    def __call__(self, s: str, env: MutableMapping[str, Any]) -> List[Dict[str, Any]]: ...



import shutil
from pathlib import Path
from typing import Any, Optional, cast

from schema_salad import codegen
from schema_salad.schema import load_schema

from .util import cwl_file_uri, get_path, metaschema_file_uri


def test_cwl_gen(tmp_path: Path) -> None:
    topmed_example_path = get_path("tests/test_real_cwl/topmed/topmed_variant_calling_pipeline.cwl")
    target_dir = tmp_path / "target"
    examples_dir = tmp_path / "examples"

    target_dir.mkdir()
    examples_dir.mkdir()
    shutil.copyfile(topmed_example_path, examples_dir / "valid_topmed.cwl")

    java_codegen(cwl_file_uri, target_dir, examples=examples_dir)
    pom_xml_path = target_dir / "pom.xml"
    assert pom_xml_path.exists()
    tests_dir = target_dir / "src" / "test" / "java" / "org" / "w3id" / "cwl" / "cwl" / "utils"
    assert tests_dir.exists()
    with open(tests_dir / "ExamplesTest.java") as f:
        assert "topmed" in f.read()


def test_meta_schema_gen(tmp_path: Path) -> None:
    target_dir = tmp_path / "target"
    target_dir.mkdir()
    java_codegen(metaschema_file_uri, target_dir)
    pom_xml_path = target_dir / "pom.xml"
    assert pom_xml_path.exists()
    src_dir = target_dir / "src" / "main" / "java" / "org" / "w3id" / "cwl" / "salad"
    assert src_dir.exists()


def java_codegen(file_uri: str, target: Path, examples: Optional[Path] = None) -> None:
    document_loader, avsc_names, schema_metadata, metaschema_loader = load_schema(file_uri)
    schema_raw_doc = metaschema_loader.fetch(file_uri)
    schema_doc, schema_metadata = metaschema_loader.resolve_all(schema_raw_doc, file_uri)
    codegen.codegen(
        "java",
        cast(list[dict[str, Any]], schema_doc),
        schema_metadata,
        document_loader,
        target=str(target),
        examples=str(examples) if examples else None,
    )



from typing import Any, Dict, Match

from _typeshed import Incomplete

from ..block_parser import BlockParser as BlockParser
from ..core import BaseRenderer as BaseRenderer
from ..core import BlockState as BlockState
from ..markdown import Markdown as Markdown
from ..toc import normalize_toc_item as normalize_toc_item
from ..toc import render_toc_ul as render_toc_ul
from ._base import BaseDirective as BaseDirective
from ._base import DirectivePlugin as DirectivePlugin

class TableOfContents(DirectivePlugin):
    min_level: Incomplete
    max_level: Incomplete
    def __init__(self, min_level: int = 1, max_level: int = 3) -> None: ...
    def generate_heading_id(self, token: Dict[str, Any], index: int) -> str: ...
    def parse(self, block: BlockParser, m: Match[str], state: BlockState) -> Dict[str, Any]: ...
    def toc_hook(self, md: Markdown, state: BlockState) -> None: ...
    def __call__(self, directive: BaseDirective, md: Markdown) -> None: ...

def render_html_toc(
    renderer: BaseRenderer, title: str, collapse: bool = False, **attrs: Any
) -> str: ...



from typing import Any, Callable, Dict, Iterable, Optional, Tuple

from .core import BlockState as BlockState
from .markdown import Markdown as Markdown
from .util import striptags as striptags

def add_toc_hook(
    md: Markdown,
    min_level: int = 1,
    max_level: int = 3,
    heading_id: Optional[Callable[[Dict[str, Any], int], str]] = None,
) -> None: ...
def normalize_toc_item(md: Markdown, token: Dict[str, Any]) -> Tuple[int, str, str]: ...
def render_toc_ul(toc: Iterable[Tuple[int, str, str]]) -> str: ...



from typing import Any, Dict, Iterable

from _typeshed import Incomplete

from ..core import BaseRenderer as BaseRenderer
from ..core import BlockState as BlockState
from ..util import strip_end as strip_end
from ._list import render_list as render_list

fenced_re: Incomplete

class MarkdownRenderer(BaseRenderer):
    NAME: str
    def __call__(self, tokens: Iterable[Dict[str, Any]], state: BlockState) -> str: ...
    def render_referrences(self, state: BlockState) -> Iterable[str]: ...
    def render_children(self, token: Dict[str, Any], state: BlockState) -> str: ...
    def text(self, token: Dict[str, Any], state: BlockState) -> str: ...
    def emphasis(self, token: Dict[str, Any], state: BlockState) -> str: ...
    def strong(self, token: Dict[str, Any], state: BlockState) -> str: ...
    def link(self, token: Dict[str, Any], state: BlockState) -> str: ...
    def image(self, token: Dict[str, Any], state: BlockState) -> str: ...
    def codespan(self, token: Dict[str, Any], state: BlockState) -> str: ...
    def linebreak(self, token: Dict[str, Any], state: BlockState) -> str: ...
    def softbreak(self, token: Dict[str, Any], state: BlockState) -> str: ...
    def blank_line(self, token: Dict[str, Any], state: BlockState) -> str: ...
    def inline_html(self, token: Dict[str, Any], state: BlockState) -> str: ...
    def paragraph(self, token: Dict[str, Any], state: BlockState) -> str: ...
    def heading(self, token: Dict[str, Any], state: BlockState) -> str: ...
    def thematic_break(self, token: Dict[str, Any], state: BlockState) -> str: ...
    def block_text(self, token: Dict[str, Any], state: BlockState) -> str: ...
    def block_code(self, token: Dict[str, Any], state: BlockState) -> str: ...
    def block_quote(self, token: Dict[str, Any], state: BlockState) -> str: ...
    def block_html(self, token: Dict[str, Any], state: BlockState) -> str: ...
    def block_error(self, token: Dict[str, Any], state: BlockState) -> str: ...
    def list(self, token: Dict[str, Any], state: BlockState) -> str: ...



from typing import Any, Dict, List, Match, Optional

from ..block_parser import BlockParser
from ..core import BlockState
from ..markdown import Markdown
from ._base import BaseDirective, DirectivePlugin

__all__ = ["Image", "Figure"]

class Image(DirectivePlugin):
    NAME: str
    def parse(self, block: BlockParser, m: Match[str], state: BlockState) -> Dict[str, Any]: ...
    def __call__(self, directive: BaseDirective, md: Markdown) -> None: ...

class Figure(DirectivePlugin):
    NAME: str
    def parse_directive_content(
        self, block: BlockParser, m: Match[str], state: BlockState
    ) -> Optional[List[Dict[str, Any]]]: ...
    def parse(self, block: BlockParser, m: Match[str], state: BlockState) -> Dict[str, Any]: ...
    def __call__(self, directive: BaseDirective, md: Markdown) -> None: ...



"""Tests of helpful error messages."""

import re

import pytest

import schema_salad
import schema_salad.main
from schema_salad.avro.schema import Names
from schema_salad.exceptions import ValidationException
from schema_salad.ref_resolver import Loader
from schema_salad.schema import load_and_validate, load_schema
from schema_salad.sourceline import cmap

from .util import get_data, get_data_uri


def test_errors() -> None:
    path = get_data("tests/test_schema/CommonWorkflowLanguage.yml")
    document_loader, avsc_names, schema_metadata, metaschema_loader = load_schema(path)
    assert isinstance(avsc_names, Names)

    for t in (
        "test_schema/test1.cwl",
        "test_schema/test2.cwl",
        "test_schema/test3.cwl",
        "test_schema/test4.cwl",
        "test_schema/test5.cwl",
        "test_schema/test6.cwl",
        "test_schema/test7.cwl",
        "test_schema/test8.cwl",
        "test_schema/test9.cwl",
        "test_schema/test10.cwl",
        "test_schema/test11.cwl",
        "test_schema/test15.cwl",
    ):
        path2 = get_data("tests/" + t)
        with pytest.raises(ValidationException):
            try:
                load_and_validate(
                    document_loader,
                    avsc_names,
                    path2,
                    True,
                )
            except ValidationException as e:
                print("\n", e)
                raise


def test_error_message1() -> None:
    path = get_data("tests/test_schema/CommonWorkflowLanguage.yml")
    document_loader, avsc_names, schema_metadata, metaschema_loader = load_schema(path)
    assert isinstance(avsc_names, Names)

    t = "test_schema/test1.cwl"
    match = (
        r"""^.+test1\.cwl:2:1: Object\s+'.+test1\.cwl'\s+is\s+not valid """
        + r"""because\s+tried 'Workflow'\s+but
\s+\* missing\s+required\s+field\s+'inputs'
\s+\* missing\s+required\s+field\s+'outputs'
\s+\* missing\s+required\s+field\s+'steps'$"""
    )
    path2 = get_data("tests/" + t)
    with pytest.raises(ValidationException, match=match):
        load_and_validate(document_loader, avsc_names, path2, True)


def test_error_message2() -> None:
    path = get_data("tests/test_schema/CommonWorkflowLanguage.yml")
    document_loader, avsc_names, schema_metadata, metaschema_loader = load_schema(path)
    assert isinstance(avsc_names, Names)

    t = "test_schema/test2.cwl"
    match = (
        r"""
^.+test2\.cwl:2:1: Field """
        r"""'class'\s+contains\s+undefined\s+reference\s+to\s+'file://.+/schema_salad/tests/test_schema/xWorkflow'$"""[
            1:
        ]
    )
    path2 = get_data("tests/" + t)
    with pytest.raises(ValidationException, match=match):
        load_and_validate(document_loader, avsc_names, path2, True)


def test_error_message3() -> None:
    path = get_data("tests/test_schema/CommonWorkflowLanguage.yml")
    document_loader, avsc_names, schema_metadata, metaschema_loader = load_schema(path)
    assert isinstance(avsc_names, Names)

    t = "test_schema/test3.cwl"
    match = r"""
^.+test3\.cwl:6:1: checking field\s+'outputs'
.+test3\.cwl:7:3:   checking object\s+'.+test3\.cwl#bar'
\s+Field 'type'\s+references\s+unknown\s+identifier\s+'xstring',\s+tried\s+file://.+/tests/test_schema/test3\.cwl#xstring$"""[  # noqa: B950
        1:
    ]
    with pytest.raises(ValidationException, match=match):
        load_and_validate(document_loader, avsc_names, get_data("tests/" + t), True)


def test_error_message4() -> None:
    path = get_data("tests/test_schema/CommonWorkflowLanguage.yml")
    document_loader, avsc_names, schema_metadata, metaschema_loader = load_schema(path)
    assert isinstance(avsc_names, Names)

    t = "test_schema/test4.cwl"
    match = r"""
^.+test4\.cwl:6:1: checking field\s+'outputs'
.+test4\.cwl:7:3:   checking object\s+'.+test4\.cwl#bar'
\s+'type'\s+field\s+is\s+int,\s+expected\s+string,\s+list,\s+or\s+a\s+dict.$"""[
        1:
    ]
    with pytest.raises(ValidationException, match=match):
        load_and_validate(document_loader, avsc_names, get_data("tests/" + t), True)


def test_error_message5() -> None:
    path = get_data("tests/test_schema/CommonWorkflowLanguage.yml")
    document_loader, avsc_names, schema_metadata, metaschema_loader = load_schema(path)
    assert isinstance(avsc_names, Names)

    t = "test_schema/test5.cwl"
    match = r"""
^.+test5\.cwl:2:1: Object\s+'.+test5\.cwl'\s+is\s+not valid because
\s+tried 'Workflow'\s+but
.+test5\.cwl:8:1:\s+the 'steps'\s+field\s+is\s+not\s+valid\s+because
\s+tried array\s+of\s+<WorkflowStep>\s+but
.+test5\.cwl:8:9:\s+item is\s+invalid\s+because
\s+is not a\s+dict.\s+Expected\s+a\s+WorkflowStep\s+object.$"""[
        1:
    ]
    with pytest.raises(ValidationException, match=match):
        load_and_validate(document_loader, avsc_names, get_data("tests/" + t), True)


def test_error_message7() -> None:
    path = get_data("tests/test_schema/CommonWorkflowLanguage.yml")
    document_loader, avsc_names, schema_metadata, metaschema_loader = load_schema(path)
    assert isinstance(avsc_names, Names)

    t = "test_schema/test7.cwl"
    match = (
        r"""^.+test7\.cwl:2:1:\s+Object\s+'.+test7\.cwl'\s+is\s+not\s+valid\s+because
\s+tried\s+'Workflow'\s+but
.+test7\.cwl:8:1:\s+the 'steps'\s+field\s+is\s+not\s+valid\s+because
\s+tried\s+array\s+of\s+<WorkflowStep>\s+but
.+test7\.cwl:9:3:\s+item is\s+invalid\s+because
\s+\* missing\s+required\s+field\s+'run'
.+test7\.cwl:10:5:\s+\* invalid\s+field\s+'scatter_method',\s+expected\s+one\s+"""
        + r"of:\s+'id',\s+'in',\s+'out',\s+'requirements',\s+'hints',\s+"
        + r"'label',\s+'doc',\s+'run',\s+'scatter',\s+'scatterMethod'$"
    )
    with pytest.raises(ValidationException, match=match):
        load_and_validate(document_loader, avsc_names, get_data("tests/" + t), True)


def test_error_message8() -> None:
    path = get_data("tests/test_schema/CommonWorkflowLanguage.yml")
    document_loader, avsc_names, schema_metadata, metaschema_loader = load_schema(path)
    assert isinstance(avsc_names, Names)

    t = "test_schema/test8.cwl"
    match = (
        r"""
^.+test8\.cwl:8:1:\s+checking field\s+'steps'
.+test8\.cwl:9:3:\s+checking object\s+'.+test8\.cwl#step1'
.+test8\.cwl:10:5:\s+"""
        r"""Field\s+'scatterMethod'\s+contains\s+undefined\s+reference\s+to\s+'file:///.+/tests/test_schema/abc'$"""[
            1:
        ]
    )
    with pytest.raises(ValidationException, match=match):
        load_and_validate(document_loader, avsc_names, get_data("tests/" + t), True)


def test_error_message9() -> None:
    path = get_data("tests/test_schema/CommonWorkflowLanguage.yml")
    document_loader, avsc_names, schema_metadata, metaschema_loader = load_schema(path)
    assert isinstance(avsc_names, Names)

    t = "test_schema/test9.cwl"
    match = (
        r"""^.+test9\.cwl:8:1:\s+checking field\s+'steps'
.+test9\.cwl:9:3:\s+checking object\s+'.+test9\.cwl#step1'
.+test9\.cwl:10:5:\s+'scatterMethod'\s+field\s+is\s+"""
        + r"""int,\s+expected\s+string,\s+list,\s+or a\s+dict.$"""
    )
    with pytest.raises(ValidationException, match=match):
        load_and_validate(document_loader, avsc_names, get_data("tests/" + t), True)


def test_error_message10() -> None:
    path = get_data("tests/test_schema/CommonWorkflowLanguage.yml")
    document_loader, avsc_names, schema_metadata, metaschema_loader = load_schema(path)
    assert isinstance(avsc_names, Names)

    t = "test_schema/test10.cwl"
    match = r"""
^.+test10\.cwl:2:1:\s+Object\s+'.+test10\.cwl'\s+is\s+not\s+valid\s+because
\s+tried 'Workflow'\s+but
.+test10\.cwl:8:1:\s+the 'steps'\s+field\s+is\s+not\s+valid\s+because
\s+tried array\s+of\s+<WorkflowStep>\s+but
.+test10\.cwl:9:3:\s+item is\s+invalid\s+because
\s+\* missing\s+required\s+field\s+'run'
.+test10\.cwl:10:5:\s+\* the\s+'scatterMethod'\s+field\s+is\s+not\s+valid\s+because
\s+value\s+is\s+a\s+CommentedSeq,\s+expected\s+null\s+or\s+ScatterMethod$"""[
        1:
    ]
    with pytest.raises(ValidationException, match=match):
        load_and_validate(document_loader, avsc_names, get_data("tests/" + t), True)


def test_error_message11() -> None:
    path = get_data("tests/test_schema/CommonWorkflowLanguage.yml")
    document_loader, avsc_names, schema_metadata, metaschema_loader = load_schema(path)
    assert isinstance(avsc_names, Names)

    t = "test_schema/test11.cwl"
    match = (
        r"""
^.+test11\.cwl:8:1:\s+checking field\s+'steps'
.+test11\.cwl:9:3:\s+checking object\s+'.+test11\.cwl#step1'
.+test11\.cwl:10:5:\s+"""
        r"""Field 'run'\s+contains\s+undefined\s+reference\s+to\s+'file://.+/tests/test_schema/blub\.cwl'$"""[
            1:
        ]
    )
    with pytest.raises(ValidationException, match=match):
        load_and_validate(document_loader, avsc_names, get_data("tests/" + t), True)


def test_error_message15() -> None:
    path = get_data("tests/test_schema/CommonWorkflowLanguage.yml")
    document_loader, avsc_names, schema_metadata, metaschema_loader = load_schema(path)
    assert isinstance(avsc_names, Names)

    t = "test_schema/test15.cwl"
    match = (
        r"""^.+test15\.cwl:3:1:\s+Object\s+'.+test15\.cwl'\s+is\s+not\s+valid\s+because
\s+tried\s+'CommandLineTool'\s+but
.+test15\.cwl:6:1:\s+the\s+'inputs'\s+field\s+is\s+not\s+valid\s+because
.+test15\.cwl:7:3:\s+item\s+is\s+invalid\s+because
.+test15\.cwl:9:5:\s+the\s+'inputBinding'\s+field\s+is\s+not\s+valid\s+because
.+tried\s+CommandLineBinding\s+but
.+test15\.cwl:11:7:             \*\s+invalid\s+field\s+'invalid_field',\s+expected\s+"""
        + r"""one\s+of:\s+'loadContents',\s+'position',\s+'prefix',\s+'separate',"""
        + r"""\s+'itemSeparator',\s+'valueFrom',\s+'shellQuote'
.+test15\.cwl:12:7:             \*\s+invalid\s+field\s+'another_invalid_field',"""
        + r"""\s+expected\s+one\s+of:\s+'loadContents',\s+'position',\s+'prefix',"""
        + r"""\s+'separate',\s+'itemSeparator',\s+'valueFrom',\s+'shellQuote'$"""
    )
    with pytest.raises(ValidationException, match=match):
        load_and_validate(document_loader, avsc_names, get_data("tests/" + t), True)


@pytest.mark.skip(
    "See https://github.com/common-workflow-language/common-workflow-language/issues/734"  # noqa: B950
)
def test_errors_previously_defined_dict_key() -> None:
    path = get_data("tests/test_schema/CommonWorkflowLanguage.yml")
    document_loader, avsc_names, schema_metadata, metaschema_loader = load_schema(path)
    assert isinstance(avsc_names, Names)

    for t in (
        "test_schema/test12.cwl",
        "test_schema/test13.cwl",
        "test_schema/test14.cwl",
    ):
        with pytest.raises(ValidationException):
            try:
                load_and_validate(
                    document_loader,
                    avsc_names,
                    get_data("tests/" + t),
                    True,
                )
            except ValidationException as e:
                print("\n", e)
                raise


def test_bad_schema() -> None:
    path = get_data("tests/bad_schema.yml")
    assert 1 == schema_salad.main.main(argsl=[path])
    path2 = get_data("tests/bad_schema.yml")
    assert 1 == schema_salad.main.main(argsl=["--print-avro", path2])


def test_bad_schema2() -> None:
    path = get_data("tests/bad_schema2.yml")
    assert 1 == schema_salad.main.main(argsl=[path])


def test_namespaces_type() -> None:
    """Confirm helpful error message when $namespaces is the wrong type."""
    with pytest.raises(
        ValidationException,
        match=re.escape("test:1:1: $namespaces must be a dictionary"),
    ):
        ldr, _, _, _ = schema_salad.schema.load_schema(
            cmap(
                {
                    "$base": "Y",
                    "name": "X",
                    "$namespaces": ["http://example.com/foo#"],
                    "$graph": [
                        {
                            "name": "ExampleType",
                            "type": "enum",
                            "symbols": ["asym", "bsym"],
                        }
                    ],
                }
            )
        )


def test_namespaces_undeclared(caplog: pytest.LogCaptureFixture) -> None:
    """Confirm warning message a namespace is used but not declared."""

    ldr, _, _, _ = schema_salad.schema.load_schema(
        cmap(
            {
                "$base": "Y",
                "name": "X",
                "$graph": [
                    {
                        "name": "namesp:ExampleType",
                        "type": "enum",
                        "symbols": ["asym", "bsym"],
                    }
                ],
            }
        )
    )

    match = (
        r""".*URI prefix 'namesp' of 'namesp:ExampleType' not recognized, """
        r"""are you missing a \$namespaces section?.*"""
    )

    assert re.match(match, caplog.text)


def test_not_a_namespace1(caplog: pytest.LogCaptureFixture) -> None:
    """Confirm no warning when relative id contains a colon but prefix doesn't look like a namespace."""

    ldr, _, _, _ = schema_salad.schema.load_schema(
        cmap(
            {
                "$base": "Y",
                "name": "X",
                "$graph": [
                    {
                        "name": "foo/colon:ExampleType",
                        "type": "enum",
                        "symbols": ["asym", "bsym"],
                    }
                ],
            }
        )
    )

    assert caplog.text == ""


def test_not_a_namespace2(caplog: pytest.LogCaptureFixture) -> None:
    """Confirm no warning when relative id contains a colon but prefix doesn't look like a namespace."""

    ldr, _, _, _ = schema_salad.schema.load_schema(
        cmap(
            {
                "$base": "Y",
                "name": "X",
                "$graph": [
                    {
                        "name": "foo#colon:ExampleType",
                        "type": "enum",
                        "symbols": ["asym", "bsym"],
                    }
                ],
            }
        )
    )

    assert caplog.text == ""


def test_not_a_namespace3(caplog: pytest.LogCaptureFixture) -> None:
    """Confirm no warning when relative id starts with a colon."""

    ldr, _, _, _ = schema_salad.schema.load_schema(
        cmap(
            {
                "$base": "Y",
                "name": "X",
                "$graph": [
                    {
                        "name": ":ExampleType",
                        "type": "enum",
                        "symbols": ["asym", "bsym"],
                    }
                ],
            }
        )
    )

    assert caplog.text == ""


def test_schemas_type() -> None:
    """Confirm helpful error message when $schemas is the wrong type."""
    path_uri = get_data_uri("tests/EDAM.owl")
    with pytest.raises(
        ValidationException,
        match=re.escape("test:1:1: $schemas must be a string or a list of string"),
    ):
        ra, _ = Loader({}).resolve_all(
            cmap(
                {
                    "$schemas": {"wrong": path_uri},
                    "edam:has_format": "edam:format_1915",
                }
            ),
            "",
        )



"""Test C++ code generation."""

import filecmp
import os
from pathlib import Path
from typing import Any, Optional, cast

import pytest

from schema_salad import codegen
from schema_salad.avro.schema import Names
from schema_salad.schema import load_schema

from .util import cwl_file_uri, get_data_uri, get_path


def test_cwl_cpp_gen(tmp_path: Path) -> None:
    """End to end test of C++ generator using the CWL v1.0 schema."""
    src_target = tmp_path / "cwl_v1_0.h"
    cpp_codegen(cwl_file_uri, src_target)
    assert os.path.exists(src_target)


@pytest.mark.parametrize(
    "filename",
    [
        "01_single_record.yml",
        "02_two_records.yml",
        "03_simple_inheritance.yml",
        "04_abstract_inheritance.yml",
        "05_specialization.yml",
    ],
)
def test_cwl_cpp_generations(tmp_path: Path, filename: str) -> None:
    """End to end test of C++ generator using small scenarios."""

    file = get_path(f"cpp_tests/{filename}")

    # file with generated cpp output
    src_target = tmp_path / "test.h"
    # file with expected cpp output
    expected = file.with_suffix(".h")

    cpp_codegen("file://" + os.fspath(file), src_target)

    assert os.path.isfile(expected)
    assert os.path.isfile(src_target)
    assert filecmp.cmp(expected, src_target, shallow=False)


def test_cwl_cpp_generations_with_spdx(tmp_path: Path) -> None:
    """End to end test of C++ generator checking for SPDX headers"""

    src_target = tmp_path / "test.h"

    input_file_uri = get_data_uri("cpp_tests/01_single_record.yml")

    """Generating different combinations of license headers"""
    """Generate License Identifier"""
    cpp_codegen(input_file_uri, src_target, spdx_license_identifier="Apache-2.0")
    lines = open(src_target).readlines()[0:2]
    assert lines[0] == "// SPDX-License-Identifier: Apache-2.0\n"
    assert lines[1] == "#pragma once\n"

    """Generate single CopyrightText"""
    cpp_codegen(
        input_file_uri,
        src_target,
        spdx_copyright_text=["Copyright 2016 Some People <email@example.com>"],
    )
    lines = open(src_target).readlines()[0:2]
    assert lines[0] == "// SPDX-FileCopyrightText: Copyright 2016 Some People <email@example.com>\n"
    assert lines[1] == "#pragma once\n"

    """Generate two CopyrightText entries"""
    cpp_codegen(
        input_file_uri,
        src_target,
        spdx_copyright_text=[
            "Copyright 2016 Person A <person_a@example.com>",
            "Copyright 2017 Person B <person_b@example.com>",
        ],
    )
    lines = open(src_target).readlines()[0:3]
    assert lines[0] == "// SPDX-FileCopyrightText: Copyright 2016 Person A <person_a@example.com>\n"
    assert lines[1] == "// SPDX-FileCopyrightText: Copyright 2017 Person B <person_b@example.com>\n"
    assert lines[2] == "#pragma once\n"

    """Generate CopyrightText and License Identifier"""
    cpp_codegen(
        input_file_uri,
        src_target,
        spdx_license_identifier="Apache-2.0",
        spdx_copyright_text=[
            "Copyright 2016 Person A <person_a@example.com>",
            "Copyright 2017 Person B <person_b@example.com>",
        ],
    )
    lines = open(src_target).readlines()[0:4]

    assert lines[0] == "// SPDX-FileCopyrightText: Copyright 2016 Person A <person_a@example.com>\n"
    assert lines[1] == "// SPDX-FileCopyrightText: Copyright 2017 Person B <person_b@example.com>\n"
    assert lines[2] == "// SPDX-License-Identifier: Apache-2.0\n"
    assert lines[3] == "#pragma once\n"


def cpp_codegen(
    file_uri: str,
    target: Path,
    spdx_copyright_text: Optional[list[str]] = None,
    spdx_license_identifier: Optional[str] = None,
) -> None:
    """Help using the C++ code generation function."""
    document_loader, avsc_names, schema_metadata, metaschema_loader = load_schema(file_uri)
    assert isinstance(avsc_names, Names)
    schema_raw_doc = metaschema_loader.fetch(file_uri)
    schema_doc, schema_metadata = metaschema_loader.resolve_all(schema_raw_doc, file_uri)
    codegen.codegen(
        "cpp",
        cast(list[dict[str, Any]], schema_doc),
        schema_metadata,
        document_loader,
        target=str(target),
        spdx_copyright_text=spdx_copyright_text,
        spdx_license_identifier=spdx_license_identifier,
    )



from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

from _typeshed import Incomplete

from .block_parser import BlockParser as BlockParser
from .core import BaseRenderer as BaseRenderer
from .core import BlockState as BlockState
from .inline_parser import InlineParser as InlineParser
from .plugins import Plugin as Plugin

class Markdown:
    renderer: Incomplete
    block: Incomplete
    inline: Incomplete
    before_parse_hooks: Incomplete
    before_render_hooks: Incomplete
    after_render_hooks: Incomplete
    def __init__(
        self,
        renderer: Optional[BaseRenderer] = None,
        block: Optional[BlockParser] = None,
        inline: Optional[InlineParser] = None,
        plugins: Optional[Iterable[Plugin]] = None,
    ) -> None: ...
    def use(self, plugin: Plugin) -> None: ...
    def render_state(self, state: BlockState) -> Union[str, List[Dict[str, Any]]]: ...
    def parse(
        self, s: str, state: Optional[BlockState] = None
    ) -> Tuple[Union[str, List[Dict[str, Any]]], BlockState]: ...
    def read(
        self, filepath: str, encoding: str = "utf-8", state: Optional[BlockState] = None
    ) -> Tuple[Union[str, List[Dict[str, Any]]], BlockState]: ...
    def __call__(self, s: str) -> Union[str, List[Dict[str, Any]]]: ...



from ..markdown import Markdown

__all__ = ["table", "table_in_quote", "table_in_list"]

def table(md: Markdown) -> None: ...
def table_in_quote(md: Markdown) -> None: ...
def table_in_list(md: Markdown) -> None: ...



from typing import ClassVar, Dict, Iterable, List, Match, Optional, Pattern, Tuple

from .core import BlockState as BlockState
from .core import Parser as Parser
from .helpers import BLOCK_TAGS as BLOCK_TAGS
from .helpers import HTML_ATTRIBUTES as HTML_ATTRIBUTES
from .helpers import HTML_TAGNAME as HTML_TAGNAME
from .helpers import LINK_LABEL as LINK_LABEL
from .helpers import PRE_TAGS as PRE_TAGS
from .helpers import parse_link_href as parse_link_href
from .helpers import parse_link_title as parse_link_title
from .helpers import unescape_char as unescape_char
from .list_parser import LIST_PATTERN as LIST_PATTERN
from .list_parser import parse_list as parse_list
from .util import escape_url as escape_url
from .util import expand_leading_tab as expand_leading_tab
from .util import expand_tab as expand_tab
from .util import unikey as unikey

class BlockParser(Parser[BlockState]):
    state_cls = BlockState
    BLANK_LINE: Pattern[str]
    RAW_HTML: str
    BLOCK_HTML: str
    SPECIFICATION: ClassVar[Dict[str, str]]
    DEFAULT_RULES: ClassVar[Iterable[str]]
    block_quote_rules: List[str]
    list_rules: List[str]
    max_nested_level: int
    def __init__(
        self,
        block_quote_rules: Optional[List[str]] = None,
        list_rules: Optional[List[str]] = None,
        max_nested_level: int = 6,
    ) -> None: ...
    def parse_blank_line(self, m: Match[str], state: BlockState) -> int: ...
    def parse_thematic_break(self, m: Match[str], state: BlockState) -> int: ...
    def parse_indent_code(self, m: Match[str], state: BlockState) -> int: ...
    def parse_fenced_code(self, m: Match[str], state: BlockState) -> Optional[int]: ...
    def parse_atx_heading(self, m: Match[str], state: BlockState) -> int: ...
    def parse_setex_heading(self, m: Match[str], state: BlockState) -> Optional[int]: ...
    def parse_ref_link(self, m: Match[str], state: BlockState) -> Optional[int]: ...
    def extract_block_quote(
        self, m: Match[str], state: BlockState
    ) -> Tuple[str, Optional[int]]: ...
    def parse_block_quote(self, m: Match[str], state: BlockState) -> int: ...
    def parse_list(self, m: Match[str], state: BlockState) -> int: ...
    def parse_block_html(self, m: Match[str], state: BlockState) -> Optional[int]: ...
    def parse_raw_html(self, m: Match[str], state: BlockState) -> Optional[int]: ...
    def parse(self, state: BlockState, rules: Optional[List[str]] = None) -> None: ...



import copy
import logging
import os
import pathlib
import re
import tempfile
import traceback
import urllib
import xml.sax  # nosec
from collections.abc import MutableMapping, MutableSequence
from io import StringIO
from typing import Any, Callable, Optional, Union, cast

import requests
from cachecontrol.caches import SeparateBodyFileCache
from cachecontrol.wrapper import CacheControl
from mypy_extensions import mypyc_attr
from rdflib.exceptions import ParserError
from rdflib.graph import Graph
from rdflib.namespace import OWL, RDF, RDFS
from rdflib.plugin import PluginException
from rdflib.plugins.parsers.notation3 import BadSyntax
from rdflib.util import guess_format
from ruamel.yaml.comments import CommentedMap, CommentedSeq, LineCol
from ruamel.yaml.error import MarkedYAMLError

from .exceptions import SchemaSaladException, ValidationException
from .fetcher import DefaultFetcher
from .sourceline import SourceLine, add_lc_filename, relname
from .utils import (
    AttachmentsType,
    CacheType,
    ContextType,
    FetcherCallableType,
    IdxResultType,
    IdxType,
    ResolvedRefType,
    ResolveType,
    aslist,
    onWindows,
    yaml_no_ts,
)

_logger = logging.getLogger("salad")


def file_uri(path: str, split_frag: bool = False) -> str:
    if path.startswith("file://"):
        return path
    if split_frag:
        pathsp = path.split("#", 2)
        if len(pathsp) == 2:
            frag = "#" + urllib.parse.quote(str(pathsp[1]))
        else:
            frag = ""
        urlpath = urllib.request.pathname2url(str(pathsp[0]))
    else:
        urlpath = urllib.request.pathname2url(path)
        frag = ""
    if urlpath.startswith("//"):
        return f"file:{urlpath}{frag}"
    return f"file://{urlpath}{frag}"


def uri_file_path(url: str) -> str:
    split = urllib.parse.urlsplit(url)
    if split.scheme == "file":
        return urllib.request.url2pathname(str(split.path)) + (
            "#" + urllib.parse.unquote(str(split.fragment)) if bool(split.fragment) else ""
        )
    raise ValidationException(f"Not a file URI: {url}")


def to_validation_exception(e: MarkedYAMLError) -> ValidationException:
    """Convert ruamel.yaml exception to our type."""
    fname_regex = re.compile(r"^file://" + re.escape(os.getcwd()) + "/")

    exc = ValidationException(e.problem)
    mark = e.problem_mark
    exc.file = re.sub(fname_regex, "", mark.name)
    exc.start = (mark.line + 1, mark.column + 1)
    exc.end = None

    if e.context:
        parent = ValidationException(e.context)
        context_mark = e.context_mark
        if context_mark:
            parent.file = re.sub(fname_regex, "", context_mark.name)
            parent.start = (context_mark.line + 1, context_mark.column + 1)
        parent.end = None
        parent.children = [exc]
        return parent
    return exc


class NormDict(dict[str, Union[CommentedMap, CommentedSeq, str, None]]):
    """A Dict where all keys are normalized using the provided function."""

    def __init__(self, normalize: Callable[[str], str] = str) -> None:
        super().__init__()
        self.normalize = normalize

    def __eq__(self, other: Any) -> bool:
        return super().__eq__(other)

    def __getitem__(self, key: Any) -> Any:
        return super().__getitem__(self.normalize(key))

    def __setitem__(self, key: Any, value: Any) -> Any:
        return super().__setitem__(self.normalize(key), value)

    def __delitem__(self, key: Any) -> Any:
        return super().__delitem__(self.normalize(key))

    def __contains__(self, key: Any) -> bool:
        return super().__contains__(self.normalize(key))

    def __del__(self) -> None:
        del self.normalize


def SubLoader(loader: "Loader") -> "Loader":
    return Loader(
        loader.ctx,
        schemagraph=loader.graph,
        foreign_properties=loader.foreign_properties,
        idx=loader.idx,
        cache=loader.cache,
        fetcher_constructor=loader.fetcher_constructor,
        skip_schemas=loader.skip_schemas,
        url_fields=loader.url_fields,
        allow_attachments=loader.allow_attachments,
        session=loader.session,
        salad_version=loader.salad_version,
    )


def _url_norm(url: str) -> str:
    return urllib.parse.urlsplit(url).geturl()


@mypyc_attr(allow_interpreted_subclasses=True)
class Loader:
    def __init__(
        self,
        ctx: ContextType,
        schemagraph: Optional[Graph] = None,
        foreign_properties: Optional[set[str]] = None,
        idx: Optional[IdxType] = None,
        cache: Optional[CacheType] = None,
        session: Optional[requests.sessions.Session] = None,
        fetcher_constructor: Optional[FetcherCallableType] = None,
        skip_schemas: Optional[bool] = None,
        url_fields: Optional[set[str]] = None,
        allow_attachments: Optional[AttachmentsType] = None,
        doc_cache: Union[str, bool] = True,
        salad_version: Optional[str] = None,
    ) -> None:
        self.idx: IdxType = NormDict(_url_norm) if idx is None else idx

        self.ctx: ContextType = {}
        self.graph = schemagraph if schemagraph is not None else Graph()
        self.foreign_properties = (
            set(foreign_properties) if foreign_properties is not None else set()
        )
        self.cache = cache if cache is not None else {}
        self.skip_schemas = skip_schemas if skip_schemas is not None else False

        if session is None:
            if doc_cache is False:
                self.session = requests.Session()
            elif doc_cache is True:
                root = pathlib.Path(os.environ.get("HOME", tempfile.gettempdir()))
                self.session = CacheControl(
                    requests.Session(),
                    cache=SeparateBodyFileCache(root / ".cache" / "salad"),
                )
            elif isinstance(doc_cache, str):
                self.session = CacheControl(
                    requests.Session(), cache=SeparateBodyFileCache(doc_cache)
                )
        else:
            self.session = session

        self.fetcher_constructor = (
            fetcher_constructor if fetcher_constructor is not None else DefaultFetcher
        )
        self.fetcher = self.fetcher_constructor(self.cache, self.session)
        self.fetch_text = self.fetcher.fetch_text
        self.check_exists = self.fetcher.check_exists
        self.url_fields: set[str] = set() if url_fields is None else set(url_fields)
        self.scoped_ref_fields: dict[str, int] = {}
        self.vocab_fields: set[str] = set()
        self.identifiers: list[str] = []
        self.identity_links: set[str] = set()
        self.standalone: Optional[set[str]] = None
        self.nolinkcheck: set[str] = set()
        self.vocab: dict[str, str] = {}
        self.rvocab: dict[str, str] = {}
        self.idmap: dict[str, str] = {}
        self.mapPredicate: dict[str, str] = {}
        self.type_dsl_fields: set[str] = set()
        self.subscopes: dict[str, str] = {}
        self.secondaryFile_dsl_fields: set[str] = set()
        self.allow_attachments = allow_attachments

        if salad_version:
            self.salad_version = salad_version
        else:
            self.salad_version = "v1.1"

        self.add_context(ctx)

    def expand_url(
        self,
        url: str,
        base_url: str,
        scoped_id: bool = False,
        vocab_term: bool = False,
        scoped_ref: Optional[int] = None,
    ) -> str:
        if url in ("@id", "@type"):
            return url

        if vocab_term and url in self.vocab:
            return url

        if url.startswith("_:"):
            return url

        if bool(self.vocab) and ":" in url:
            prefix = url.split(":")[0]
            if not prefix:
                pass
            elif prefix in self.vocab:
                url = self.vocab[prefix] + url[len(prefix) + 1 :]
            elif (
                prefix not in self.fetcher.supported_schemes()
                and "/" not in prefix
                and "#" not in prefix
            ):
                _logger.warning(
                    "URI prefix '%s' of '%s' not recognized, are you missing a "
                    "$namespaces section?",
                    prefix,
                    url,
                )

        split = urllib.parse.urlsplit(url)

        if (
            (bool(split.scheme) and split.scheme in self.fetcher.supported_schemes())
            or url.startswith("$(")
            or url.startswith("${")
        ):
            pass
        elif scoped_id and not bool(split.fragment):
            splitbase = urllib.parse.urlsplit(base_url)
            frg = splitbase.fragment + "/" + split.path if bool(splitbase.fragment) else split.path
            pt = splitbase.path if splitbase.path != "" else "/"
            url = urllib.parse.urlunsplit(
                (splitbase.scheme, splitbase.netloc, pt, splitbase.query, frg)
            )
        elif scoped_ref is not None and not split.fragment:
            pass
        else:
            url = self.fetcher.urljoin(base_url, url)

        if vocab_term and url in self.rvocab:
            return self.rvocab[url]
        return url

    def _add_properties(self, s: str) -> None:
        for _, _, rng in self.graph.triples((s, RDFS.range, None)):
            literal = (
                str(rng).startswith("http://www.w3.org/2001/XMLSchema#")
                and str(rng) != "http://www.w3.org/2001/XMLSchema#anyURI"
            ) or str(rng) == "http://www.w3.org/2000/01/rdf-schema#Literal"
            if not literal:
                self.url_fields.add(str(s))
        self.foreign_properties.add(str(s))

    def add_namespaces(self, ns: dict[str, str]) -> None:
        """Add the given namespace to our vocab list."""
        self.vocab.update(ns)

    def add_schemas(self, ns: Union[list[str], str], base_url: str) -> None:
        """Fetch exernal schemas and add them to the graph."""
        if self.skip_schemas:
            return
        for sch in aslist(ns):
            fetchurl = self.fetcher.urljoin(base_url, sch)
            if fetchurl not in self.cache or self.cache[fetchurl] is True:
                _logger.debug("Getting external schema %s", fetchurl)
                try:
                    content = self.fetch_text(fetchurl)
                except Exception as e:
                    tb = traceback.format_exception(type(e), e, e.__traceback__)
                    _logger.warning("Could not load extension schema %s: %s", fetchurl, str(e))
                    _logger.debug(tb)
                    continue
                newGraph = Graph()
                err_msg = "unknown error"
                for fmt in [
                    guess_format(sch),
                    "xml",
                    "turtle",
                    None,
                    guess_format(sch),
                ]:
                    try:
                        newGraph.parse(data=content, format=fmt, publicID=str(fetchurl))
                        self.cache[fetchurl] = newGraph
                        self.graph += newGraph
                        break
                    except (
                        xml.sax.SAXParseException,
                        TypeError,
                        BadSyntax,
                        ParserError,
                        PluginException,
                    ) as e:
                        err_msg = str(e)
                else:
                    _logger.warning("Could not load extension schema %s: %s", fetchurl, err_msg)

        for s, _, _ in self.graph.triples((None, RDF.type, RDF.Property)):
            self._add_properties(s)
        for s, _, o in self.graph.triples((None, RDFS.subPropertyOf, None)):
            self._add_properties(s)
            self._add_properties(o)
        for s, _, _ in self.graph.triples((None, RDFS.range, None)):
            self._add_properties(s)
        for s, _, _ in self.graph.triples((None, RDF.type, OWL.ObjectProperty)):
            self._add_properties(s)

        for s, _, _ in self.graph.triples((None, None, None)):
            self.idx[str(s)] = None

    def add_context(self, newcontext: ContextType) -> None:
        if bool(self.vocab):
            raise ValidationException("Refreshing context that already has stuff in it")

        self.url_fields = {"$schemas"}
        self.scoped_ref_fields.clear()
        self.vocab_fields.clear()
        self.identifiers.clear()
        self.identity_links.clear()
        self.standalone = set()
        self.nolinkcheck.clear()
        self.idmap.clear()
        self.mapPredicate.clear()
        self.vocab.clear()
        self.rvocab.clear()
        self.type_dsl_fields.clear()
        self.secondaryFile_dsl_fields.clear()
        self.subscopes.clear()

        self.ctx.update(_copy_dict_without_key(newcontext, "@context"))

        _logger.debug("ctx is %s", self.ctx)

        for key, value in self.ctx.items():
            if value == "@id":
                self.identifiers.append(key)
                self.identity_links.add(key)
            elif isinstance(value, MutableMapping):
                if value.get("@type") == "@id":
                    self.url_fields.add(key)
                    if "refScope" in value:
                        self.scoped_ref_fields[key] = value["refScope"]
                    if value.get("identity", False):
                        self.identity_links.add(key)

                if value.get("@type") == "@vocab":
                    self.url_fields.add(key)
                    self.vocab_fields.add(key)
                    if "refScope" in value:
                        self.scoped_ref_fields[key] = value["refScope"]
                    if value.get("typeDSL"):
                        self.type_dsl_fields.add(key)

                if value.get("secondaryFilesDSL"):
                    self.secondaryFile_dsl_fields.add(key)

                if value.get("noLinkCheck"):
                    self.nolinkcheck.add(key)

                if value.get("mapSubject"):
                    self.idmap[key] = value["mapSubject"]

                if value.get("mapPredicate"):
                    self.mapPredicate[key] = value["mapPredicate"]

                if value.get("@id"):
                    self.vocab[key] = value["@id"]

                if value.get("subscope"):
                    self.subscopes[key] = value["subscope"]

            elif isinstance(value, str):
                self.vocab[key] = value

        for k, v in self.vocab.items():
            self.rvocab[self.expand_url(v, "", scoped_id=False)] = k

        self.identifiers.sort()

        _logger.debug("identifiers is %s", self.identifiers)
        _logger.debug("identity_links is %s", self.identity_links)
        _logger.debug("url_fields is %s", self.url_fields)
        _logger.debug("vocab_fields is %s", self.vocab_fields)
        _logger.debug("vocab is %s", self.vocab)

    def resolve_ref(
        self,
        ref: ResolveType,
        base_url: Optional[str] = None,
        checklinks: bool = True,
        strict_foreign_properties: bool = False,
        content_types: Optional[list[str]] = None,  # Expected content-types
    ) -> ResolvedRefType:
        lref = ref
        obj: Optional[CommentedMap] = None
        resolved_obj: ResolveType = None
        imp = False
        inc = False
        mixin: Optional[MutableMapping[str, str]] = None

        if not base_url:
            base_url = file_uri(os.getcwd()) + "/"

        # If `ref` is a dict, look for special directives.
        if isinstance(lref, CommentedMap):
            obj = lref
            if "$import" in obj:
                if len(obj) == 1:
                    lref = obj["$import"]
                    imp = True
                    obj = None
                else:
                    raise ValidationException(
                        f"'$import' must be the only field in {obj}",
                        SourceLine(obj, "$import"),
                    )
            elif "$include" in obj:
                if len(obj) == 1:
                    lref = obj["$include"]
                    inc = True
                    obj = None
                else:
                    raise ValidationException(
                        f"'$include' must be the only field in {obj}",
                        SourceLine(obj, "$include"),
                    )
            elif "$mixin" in obj:
                lref = obj["$mixin"]
                mixin = obj
                obj = None
            else:
                lref = None
                for identifier in self.identifiers:
                    if identifier in obj:
                        lref = obj[identifier]
                        break
                if not lref:
                    raise ValidationException(
                        f"Object {obj!r} does not have identifier field in {self.identifiers}",
                        SourceLine(obj),
                    )

        if not isinstance(lref, str):
            raise ValidationException(
                f"Expected CommentedMap or string, got {type(lref)}: {lref!r}"
            )

        if isinstance(lref, str) and os.sep == "\\":
            # Convert Windows path separator in ref
            lref = lref.replace("\\", "/")

        url = self.expand_url(lref, base_url, scoped_id=(obj is not None))
        # Has this reference been loaded already?
        if url in self.idx and (not mixin):
            resolved_obj = self.idx[url]
            if isinstance(resolved_obj, MutableMapping):
                metadata: Union[CommentedMap, CommentedSeq, str, None] = self.idx.get(
                    urllib.parse.urldefrag(url)[0], CommentedMap()
                )
                if isinstance(metadata, MutableMapping):
                    if "$graph" in resolved_obj:
                        metadata = _copy_dict_without_key(resolved_obj, "$graph")
                        return resolved_obj["$graph"], metadata
                    else:
                        return resolved_obj, metadata
                else:
                    raise ValidationException(
                        f"Expected CommentedMap, got {type(metadata)}: {metadata!r}"
                    )
            elif isinstance(resolved_obj, MutableSequence):
                metadata = self.idx.get(urllib.parse.urldefrag(url)[0], CommentedMap())
                if isinstance(metadata, MutableMapping):
                    return resolved_obj, metadata
                else:
                    return resolved_obj, CommentedMap()
            elif isinstance(resolved_obj, str):
                return resolved_obj, CommentedMap()
            else:
                raise ValidationException(
                    f"Expected MutableMapping or MutableSequence, got "
                    f"{type(resolved_obj)}: {resolved_obj!r}"
                )

        # "$include" directive means load raw text
        if inc:
            # Make a note in the index that this was an included string
            self.idx["include:" + url] = url
            return self.fetch_text(url), CommentedMap()

        doc = None
        if isinstance(obj, MutableMapping):
            for identifier in self.identifiers:
                obj[identifier] = url
            doc_url = url
        else:
            # Load structured document
            doc_url, frg = urllib.parse.urldefrag(url)
            if doc_url in self.idx and (not mixin):
                # If the base document is in the index, it was already loaded,
                # so if we didn't find the reference earlier then it must not
                # exist.
                raise SourceLine(self.idx, doc_url, ValidationException).makeError(
                    f"Reference '#{frg}' not found in file {doc_url!r}.",
                )
            doc = self.fetch(doc_url, inject_ids=(not mixin), content_types=content_types)

        if imp:
            # Make a note in the index that this was an imported fragment
            self.idx["import:" + url] = url

        # Recursively expand urls and resolve directives
        if bool(mixin):
            doc = copy.deepcopy(doc)
            if isinstance(doc, CommentedMap) and mixin is not None:
                doc.update(mixin)
                del doc["$mixin"]
            resolved_obj, metadata = self.resolve_all(
                doc,
                base_url,
                file_base=doc_url,
                checklinks=checklinks,
                strict_foreign_properties=strict_foreign_properties,
            )
        else:
            resolved_obj, metadata = self.resolve_all(
                doc or obj,
                doc_url,
                checklinks=checklinks,
                strict_foreign_properties=strict_foreign_properties,
            )

        # Requested reference should be in the index now, otherwise it's a bad
        # reference
        if not bool(mixin):
            if url in self.idx:
                resolved_obj = self.idx[url]
            else:
                raise ValidationException(
                    f"Reference {url!r} is not in the index. Index contains: {' '.join(self.idx)}"
                )

        if isinstance(resolved_obj, CommentedMap):
            if "$graph" in resolved_obj:
                metadata = _copy_dict_without_key(resolved_obj, "$graph")
                return resolved_obj["$graph"], metadata
            else:
                return resolved_obj, metadata
        else:
            return resolved_obj, metadata

    def _resolve_idmap(
        self,
        document: CommentedMap,
        loader: "Loader",
    ) -> None:
        # Convert fields with mapSubject into lists
        # use mapPredicate if the mapped value isn't a dict.
        for idmapField in loader.idmap:
            if idmapField in document:
                idmapFieldValue = document[idmapField]
                if (
                    isinstance(idmapFieldValue, MutableMapping)
                    and "$import" not in idmapFieldValue
                    and "$include" not in idmapFieldValue
                ):
                    ls = CommentedSeq()
                    for k in sorted(idmapFieldValue.keys()):
                        val = idmapFieldValue[k]
                        v: Optional[CommentedMap] = None
                        if not isinstance(val, CommentedMap):
                            if idmapField in loader.mapPredicate:
                                v = CommentedMap(((loader.mapPredicate[idmapField], val),))
                                v.lc.add_kv_line_col(
                                    loader.mapPredicate[idmapField],
                                    document[idmapField].lc.data[k],
                                )
                                v.lc.filename = document.lc.filename
                            else:
                                raise ValidationException(
                                    f"mapSubject {k!r} value {v!r} is not a dict "
                                    "and does not have a mapPredicate.",
                                    SourceLine(document, idmapField),
                                )
                        else:
                            v = val

                        v[loader.idmap[idmapField]] = k
                        v.lc.add_kv_line_col(
                            loader.idmap[idmapField], document[idmapField].lc.data[k]
                        )
                        v.lc.filename = document.lc.filename

                        ls.lc.add_kv_line_col(len(ls), document[idmapField].lc.data[k])

                        ls.lc.filename = document.lc.filename
                        ls.append(v)

                    document[idmapField] = ls

    def _type_dsl(
        self,
        t: Union[str, CommentedMap, CommentedSeq],
        lc: LineCol,
        filename: str,
    ) -> Union[str, CommentedMap, CommentedSeq]:
        if not isinstance(t, str):
            return t

        t_ = t
        optional = False
        if t_.endswith("?"):
            optional = True
            t_ = t_[0:-1]

        if t_.endswith("[]"):
            salad_versions = [int(v) for v in self.salad_version[1:].split(".")]
            rest = t_[0:-2]
            if salad_versions < [1, 3]:
                if rest.endswith("[]"):
                    # To show the error message with the original type
                    return t
                else:
                    cmap = CommentedMap((("type", "array"), ("items", rest)))
            else:
                items = self._type_dsl(rest, lc, filename)
                cmap = CommentedMap((("type", "array"), ("items", items)))
            cmap.lc.add_kv_line_col("type", lc)
            cmap.lc.add_kv_line_col("items", lc)
            cmap.lc.filename = filename
            expanded: Union[str, CommentedMap, CommentedSeq] = cmap
        else:
            expanded = t_

        if optional:
            cs = CommentedSeq(["null", expanded])
            cs.lc.add_kv_line_col(0, lc)
            cs.lc.add_kv_line_col(1, lc)
            cs.lc.filename = filename
            ret: Union[str, CommentedMap, CommentedSeq] = cs
        else:
            ret = expanded

        return ret

    def _secondaryFile_dsl(
        self,
        t: Union[str, CommentedMap, CommentedSeq],
        lc: LineCol,
        filename: str,
    ) -> Union[str, CommentedMap, CommentedSeq]:
        if not isinstance(t, str):
            return t
        pat = t[0:-1] if t.endswith("?") else t
        req: Optional[bool] = False if t.endswith("?") else None

        second = CommentedMap((("pattern", pat), ("required", req)))
        second.lc.add_kv_line_col("pattern", lc)
        second.lc.add_kv_line_col("required", lc)
        second.lc.filename = filename
        return second

    def _apply_dsl(
        self,
        datum: Union[str, CommentedMap, CommentedSeq],
        d: str,
        loader: "Loader",
        lc: LineCol,
        filename: str,
    ) -> Union[str, CommentedMap, CommentedSeq]:
        if d in loader.type_dsl_fields:
            return self._type_dsl(datum, lc, filename)
        if d in loader.secondaryFile_dsl_fields:
            return self._secondaryFile_dsl(datum, lc, filename)
        return datum

    def _resolve_dsl(
        self,
        document: CommentedMap,
        loader: "Loader",
    ) -> None:
        fields = list(loader.type_dsl_fields)
        fields.extend(loader.secondaryFile_dsl_fields)

        for d in fields:
            if d in document:
                datum2 = datum = document[d]
                if isinstance(datum, str):
                    datum2 = self._apply_dsl(
                        datum,
                        d,
                        loader,
                        document.lc.data[d] if document.lc.data else document.lc,
                        getattr(document.lc, "filename", ""),
                    )
                elif isinstance(datum, CommentedSeq):
                    datum2 = CommentedSeq()
                    for n, t in enumerate(datum):
                        if datum.lc and datum.lc.data:
                            datum2.lc.add_kv_line_col(len(datum2), datum.lc.data[n])
                            datum2.append(
                                self._apply_dsl(
                                    t, d, loader, datum.lc.data[n], document.lc.filename
                                )
                            )
                        else:
                            datum2.append(self._apply_dsl(t, d, loader, LineCol(), ""))
                if isinstance(datum2, CommentedSeq):
                    datum3 = CommentedSeq()
                    seen: list[str] = []
                    for i, item in enumerate(datum2):
                        if isinstance(item, CommentedSeq):
                            for j, v in enumerate(item):
                                if v not in seen:
                                    datum3.lc.add_kv_line_col(len(datum3), item.lc.data[j])
                                    datum3.append(v)
                                    seen.append(v)
                        else:
                            if item not in seen:
                                if datum2.lc and datum2.lc.data:
                                    datum3.lc.add_kv_line_col(len(datum3), datum2.lc.data[i])
                                datum3.append(item)
                                seen.append(item)
                    document[d] = datum3
                else:
                    document[d] = datum2

    def _resolve_identifier(self, document: CommentedMap, loader: "Loader", base_url: str) -> str:
        # Expand identifier field (usually 'id') to resolve scope
        for identifier in loader.identifiers:
            if identifier in document:
                if isinstance(document[identifier], str):
                    document[identifier] = loader.expand_url(
                        document[identifier], base_url, scoped_id=True
                    )
                    if document[identifier] not in loader.idx or isinstance(
                        loader.idx[document[identifier]], str
                    ):
                        loader.idx[document[identifier]] = document
                    base_url = document[identifier]
                else:
                    raise ValidationException(
                        f"identifier field {document[identifier]!r} must be a string"
                    )
        return base_url

    def _resolve_identity(
        self,
        document: dict[str, Union[str, MutableSequence[Union[str, CommentedMap]]]],
        loader: "Loader",
        base_url: str,
    ) -> None:
        # Resolve scope for identity fields (fields where the value is the
        # identity of a standalone node, such as enum symbols)
        for identifier in loader.identity_links:
            if identifier in document and isinstance(document[identifier], MutableSequence):
                for n, v in enumerate(document[identifier]):
                    if isinstance(v, str):
                        document[identifier][n] = loader.expand_url(  # type: ignore
                            v, base_url, scoped_id=True
                        )
                        if document[identifier][n] not in loader.idx:
                            loader.idx[cast(str, document[identifier][n])] = v

    def _normalize_fields(self, document: CommentedMap, loader: "Loader") -> None:
        # Normalize fields which are prefixed or full URIn to vocabulary terms
        for d in list(document.keys()):
            if isinstance(d, str):
                d2 = loader.expand_url(d, "", scoped_id=False, vocab_term=True)
                if d != d2:
                    document[d2] = document[d]
                    document.lc.add_kv_line_col(d2, document.lc.data[d])
                    del document[d]

    def _resolve_uris(
        self,
        document: dict[str, Union[str, MutableSequence[Union[str, CommentedMap]]]],
        loader: "Loader",
        base_url: str,
    ) -> None:
        # Resolve remaining URLs based on document base
        for d in loader.url_fields:
            if d in document:
                datum = document[d]
                if isinstance(datum, str):
                    document[d] = loader.expand_url(
                        datum,
                        base_url,
                        scoped_id=False,
                        vocab_term=(d in loader.vocab_fields),
                        scoped_ref=loader.scoped_ref_fields.get(d),
                    )
                elif isinstance(datum, MutableSequence):
                    for i, url in enumerate(datum):
                        if isinstance(url, str):
                            datum[i] = loader.expand_url(
                                url,
                                base_url,
                                scoped_id=False,
                                vocab_term=(d in loader.vocab_fields),
                                scoped_ref=loader.scoped_ref_fields.get(d),
                            )

    def resolve_all(
        self,
        document: ResolveType,
        base_url: str,
        file_base: Optional[str] = None,
        checklinks: bool = True,
        strict_foreign_properties: bool = False,
    ) -> ResolvedRefType:
        loader = self
        metadata = CommentedMap()
        if file_base is None:
            file_base = base_url

        if isinstance(document, CommentedMap):
            # Handle $import and $include
            if "$import" in document or "$include" in document:
                return self.resolve_ref(
                    document,
                    base_url=file_base,
                    checklinks=checklinks,
                    strict_foreign_properties=strict_foreign_properties,
                )
            if "$mixin" in document:
                return self.resolve_ref(
                    document,
                    base_url=base_url,
                    checklinks=checklinks,
                    strict_foreign_properties=strict_foreign_properties,
                )
        elif isinstance(document, CommentedSeq):
            pass
        elif isinstance(document, (list, dict)):
            raise ValidationException(
                f"Expected CommentedMap or CommentedSeq, got {type(document)}: {document!r}"
            )
        else:
            return (document, metadata)

        newctx: Optional["Loader"] = None
        if isinstance(document, CommentedMap):
            # Handle $base, $profile, $namespaces, $schemas and $graph
            if "$base" in document:
                base_url = document["$base"]

            if "$profile" in document:
                if newctx is None:
                    newctx = SubLoader(self)
                newctx.add_namespaces(document.get("$namespaces", CommentedMap()))
                newctx.add_schemas(document.get("$schemas", []), document["$profile"])

            if "$namespaces" in document:
                if newctx is None:
                    newctx = SubLoader(self)
                namespaces = document["$namespaces"]
                if isinstance(namespaces, dict):
                    newctx.add_namespaces(document["$namespaces"])
                else:
                    raise ValidationException(
                        "$namespaces must be a dictionary",
                        SourceLine(document, "$namespaces"),
                    )

            if "$schemas" in document:
                if newctx is None:
                    newctx = SubLoader(self)
                schemas = document["$schemas"]
                if isinstance(schemas, (list, str)):
                    newctx.add_schemas(schemas, file_base)
                else:
                    raise ValidationException(
                        "$schemas must be a string or a list of string",
                        SourceLine(document, "$schemas"),
                    )

            if newctx is not None:
                loader = newctx

            for identifier in loader.identity_links:
                if identifier in document:
                    if isinstance(document[identifier], str):
                        document[identifier] = loader.expand_url(
                            document[identifier], base_url, scoped_id=True
                        )
                        loader.idx[document[identifier]] = document

            metadata = document
            if "$graph" in document:
                document = document["$graph"]

        if isinstance(document, CommentedMap):
            self._normalize_fields(document, loader)
            self._resolve_idmap(document, loader)
            self._resolve_dsl(document, loader)
            base_url = self._resolve_identifier(document, loader, base_url)
            self._resolve_identity(document, loader, base_url)
            self._resolve_uris(document, loader, base_url)

            try:
                for key, val in document.items():
                    subscope: str = ""
                    if key in loader.subscopes:
                        subscope = "/" + loader.subscopes[key]
                    document[key], _ = loader.resolve_all(
                        val, base_url + subscope, file_base=file_base, checklinks=False
                    )
            except ValidationException as v:
                _logger.warning("loader is %s", id(loader), exc_info=True)
                raise ValidationException(
                    f"({id(loader)}) ({file_base}) Validation error in field {key}:",
                    None,
                    [v],
                ) from v

        elif isinstance(document, CommentedSeq):
            i = 0
            try:
                while i < len(document):
                    val = document[i]
                    if isinstance(val, CommentedMap) and ("$import" in val or "$mixin" in val):
                        l, import_metadata = loader.resolve_ref(
                            val, base_url=file_base, checklinks=False
                        )
                        metadata.setdefault("$import_metadata", {})
                        for identifier in loader.identifiers:
                            if identifier in import_metadata:
                                metadata["$import_metadata"][
                                    import_metadata[identifier]
                                ] = import_metadata
                        if isinstance(l, CommentedSeq):
                            lc = document.lc.data[i]
                            del document[i]
                            llen = len(l)
                            for j in range(len(document) + llen, i + llen, -1):
                                document.lc.data[j - 1] = document.lc.data[j - llen]
                            for item in l:
                                document.insert(i, item)
                                document.lc.data[i] = lc
                                i += 1
                        else:
                            document[i] = l
                            i += 1
                    else:
                        document[i], _ = loader.resolve_all(
                            val, base_url, file_base=file_base, checklinks=False
                        )
                        i += 1
            except ValidationException as v:
                _logger.warning("failed", exc_info=True)
                raise ValidationException(
                    f"({id(loader)}) ({file_base}) Validation error in position {i}:",
                    None,
                    [v],
                ) from v

        if checklinks:
            all_doc_ids: dict[str, str] = {}
            loader.validate_links(
                document,
                "",
                all_doc_ids,
                strict_foreign_properties=strict_foreign_properties,
            )

        return document, metadata

    def fetch(
        self,
        url: str,
        inject_ids: bool = True,
        content_types: Optional[list[str]] = None,
    ) -> IdxResultType:
        if url in self.idx:
            return self.idx[url]
        try:
            text = self.fetch_text(url, content_types=content_types)
            textIO = StringIO(text)
            textIO.name = str(url)
            yaml = yaml_no_ts()
            attachments = yaml.load_all(textIO)
            result = cast(Union[CommentedSeq, CommentedMap], next(attachments))

            if self.allow_attachments is not None and self.allow_attachments(result):
                i = 1
                for a in attachments:
                    self.idx[f"{url}#attachment-{i}"] = a
                    i += 1
            add_lc_filename(result, url)
        except MarkedYAMLError as e:
            raise to_validation_exception(e) from e
        if isinstance(result, CommentedMap) and inject_ids and bool(self.identifiers):
            missing_identifier = True
            for identifier in self.identifiers:
                if identifier in result:
                    missing_identifier = False
                    self.idx[self.expand_url(result[identifier], url, scoped_id=True)] = result
            if missing_identifier:
                result[self.identifiers[0]] = url
        self.idx[url] = result
        return result

    def validate_scoped(self, field: str, link: str, docid: str) -> str:
        split = urllib.parse.urlsplit(docid)
        sp = split.fragment.split("/")
        n = self.scoped_ref_fields[field]
        while n > 0 and len(sp) > 0:
            sp.pop()
            n -= 1
        tried = []
        while True:
            sp.append(link)
            url = urllib.parse.urlunsplit(
                (split.scheme, split.netloc, split.path, split.query, "/".join(sp))
            )
            tried.append(url)
            if url in self.idx:
                return url
            sp.pop()
            if len(sp) == 0:
                break
            sp.pop()
        if onWindows() and link.startswith("file:"):
            link = link.lower()
        raise ValidationException(
            f"Field {field!r} references unknown identifier {link!r}, tried {' '.join(tried)}"
        )

    def validate_link(
        self,
        field: str,
        link: Union[str, CommentedSeq, CommentedMap],
        # link also can be None, but that results in
        # mypyc "error: Local variable "link" has inferred type None; add an annotation"
        docid: str,
        all_doc_ids: dict[str, str],
    ) -> Union[str, CommentedSeq, CommentedMap]:
        if field in self.nolinkcheck:
            return link
        if isinstance(link, str):
            if field in self.vocab_fields:
                if link not in self.vocab and link not in self.idx and link not in self.rvocab:
                    if field in self.scoped_ref_fields:
                        return self.validate_scoped(field, link, docid)
                    elif not self.check_exists(link):
                        raise ValidationException(
                            f"Field {field!r} contains undefined reference to {link!r}"
                        )
            elif link not in self.idx and link not in self.rvocab:
                if field in self.scoped_ref_fields:
                    return self.validate_scoped(field, link, docid)
                elif not self.check_exists(link):
                    raise ValidationException(
                        f"Field {field!r} contains undefined reference to {link!r}"
                    )
        elif isinstance(link, CommentedSeq):
            errors = []
            for n, i in enumerate(link):
                try:
                    link[n] = self.validate_link(field, i, docid, all_doc_ids)
                except ValidationException as v:
                    errors.append(v)
            if bool(errors):
                raise ValidationException("", None, errors)
        elif isinstance(link, CommentedMap):
            self.validate_links(link, docid, all_doc_ids)
        elif link is None:  # type: ignore[unreachable]
            return None
        else:
            raise ValidationException(
                f"{field!r} field is {type(link).__name__}, expected string, list, or a dict."
            )
        return link

    def getid(self, d: Any) -> Optional[str]:
        if isinstance(d, MutableMapping):
            for i in self.identifiers:
                if i in d:
                    idd = d[i]
                    if isinstance(idd, str):
                        return idd
        return None

    def validate_links(
        self,
        document: ResolveType,
        base_url: str,
        all_doc_ids: dict[str, str],
        strict_foreign_properties: bool = False,
    ) -> None:
        docid = self.getid(document) or base_url

        errors: list[SchemaSaladException] = []
        iterator: Any = None
        if isinstance(document, MutableSequence):
            iterator = enumerate(document)
        elif isinstance(document, MutableMapping):
            for d in self.url_fields:
                try:
                    if d in document and d not in self.identity_links:
                        document[d] = self.validate_link(d, document[d], docid, all_doc_ids)
                except SchemaSaladException as v:
                    v = v.with_sourceline(SourceLine(document, d, str))
                    if d == "$schemas" or (
                        d in self.foreign_properties and not strict_foreign_properties
                    ):
                        _logger.warning(v.as_warning())
                    else:
                        errors.append(v)
            # TODO: Validator should local scope only in which
            # duplicated keys are prohibited.
            # See also https://github.com/common-workflow-language/common-workflow-language/issues/734  # noqa: B950
            # In the future, it should raise
            # ValidationException instead of _logger.warn
            try:
                for identifier in self.identifiers:  # validate that each id is defined uniquely
                    if identifier in document:
                        sl = SourceLine(document, identifier, str)
                        if (
                            document[identifier] in all_doc_ids
                            and sl.makeLead() != all_doc_ids[document[identifier]]
                        ):
                            _logger.warning(
                                "%s object %s %r previously defined",
                                all_doc_ids[document[identifier]],
                                identifier,
                                relname(document[identifier]),
                            )
                        else:
                            all_doc_ids[document[identifier]] = sl.makeLead()
                            break
            except ValidationException as v:
                errors.append(v.with_sourceline(sl))

            iterator = list(document.items())
        else:
            return

        for key, val in iterator:
            sl = SourceLine(document, key, str)
            try:
                self.validate_links(
                    val,
                    docid,
                    all_doc_ids,
                    strict_foreign_properties=strict_foreign_properties,
                )
            except ValidationException as v:
                if key in self.nolinkcheck or (isinstance(key, str) and ":" in key):
                    _logger.warning(v.as_warning())
                else:
                    docid2 = self.getid(val)
                    if docid2 is not None:
                        errors.append(
                            ValidationException(f"checking object {relname(docid2)!r}", sl, [v])
                        )
                    else:
                        if isinstance(key, str):
                            errors.append(ValidationException(f"checking field {key!r}", sl, [v]))
                        else:
                            errors.append(ValidationException("checking item", sl, [v]))
        if bool(errors):
            if len(errors) > 1:
                raise ValidationException("", None, errors)
            raise errors[0]
        return


def _copy_dict_without_key(
    from_dict: Union[CommentedMap, ContextType], filtered_key: str
) -> CommentedMap:
    new_dict = CommentedMap(from_dict.items())
    if filtered_key in new_dict:
        del new_dict[filtered_key]
    if isinstance(from_dict, CommentedMap):
        new_dict.lc.data = copy.copy(from_dict.lc.data)
        new_dict.lc.filename = from_dict.lc.filename
    return new_dict



from ..markdown import Markdown

__all__ = ["math", "math_in_quote", "math_in_list"]

def math(md: Markdown) -> None: ...
def math_in_quote(md: Markdown) -> None: ...
def math_in_list(md: Markdown) -> None: ...



from typing import Any, Dict, Tuple, Union

from _typeshed import Incomplete

from .util import escape_url as escape_url

PREVENT_BACKSLASH: str
PUNCTUATION: Incomplete
LINK_LABEL: str
LINK_BRACKET_START: Incomplete
LINK_BRACKET_RE: Incomplete
LINK_HREF_BLOCK_RE: Incomplete
LINK_HREF_INLINE_RE: Incomplete
LINK_TITLE_RE: Incomplete
PAREN_END_RE: Incomplete
HTML_TAGNAME: str
HTML_ATTRIBUTES: str
BLOCK_TAGS: Incomplete
PRE_TAGS: Incomplete

def unescape_char(text: str) -> str: ...
def parse_link_text(src: str, pos: int) -> Union[Tuple[str, int], Tuple[None, None]]: ...
def parse_link_label(src: str, start_pos: int) -> Union[Tuple[str, int], Tuple[None, None]]: ...
def parse_link_href(
    src: str, start_pos: int, block: bool = False
) -> Union[Tuple[str, int], Tuple[None, None]]: ...
def parse_link_title(
    src: str, start_pos: int, max_pos: int
) -> Union[Tuple[str, int], Tuple[None, None]]: ...
def parse_link(src: str, pos: int) -> Union[Tuple[Dict[str, Any], int], Tuple[None, None]]: ...



"""Generate language specific loaders for a particular SALAD schema."""

import sys
from collections.abc import MutableMapping, MutableSequence
from io import TextIOWrapper
from typing import Any, Optional, TextIO, Union
from urllib.parse import urlsplit

from . import schema
from .codegen_base import CodeGenBase
from .cpp_codegen import CppCodeGen
from .dlang_codegen import DlangCodeGen
from .dotnet_codegen import DotNetCodeGen
from .exceptions import SchemaSaladException
from .java_codegen import JavaCodeGen
from .python_codegen import PythonCodeGen
from .ref_resolver import Loader
from .schema import shortname
from .typescript_codegen import TypeScriptCodeGen
from .utils import aslist

FIELD_SORT_ORDER = ["class", "id", "name"]


def codegen(
    lang: str,
    i: list[dict[str, str]],
    schema_metadata: dict[str, Any],
    loader: Loader,
    target: Optional[str] = None,
    examples: Optional[str] = None,
    package: Optional[str] = None,
    copyright: Optional[str] = None,
    spdx_copyright_text: Optional[list[str]] = None,
    spdx_license_identifier: Optional[str] = None,
    parser_info: Optional[str] = None,
) -> None:
    """Generate classes with loaders for the given Schema Salad description."""
    j = schema.extend_and_specialize(i, loader)

    gen: Optional[CodeGenBase] = None
    base = schema_metadata.get("$base", schema_metadata.get("id"))
    # ``urlsplit`` decides whether to return an encoded result based
    # on the object type. To ensure the code behaves the same for Py
    # 3.6+, we enforce that the input value is of type ``str``.
    if base is None:
        base = ""
    sp = urlsplit(base)
    pkg = (
        package
        if package
        else ".".join(list(reversed(sp.netloc.split("."))) + sp.path.strip("/").split("/"))
    )
    info = parser_info or pkg
    salad_version = schema_metadata.get("saladVersion", "v1.1")

    if lang in {"python", "cpp", "dlang"}:
        if target:
            dest: Union[TextIOWrapper, TextIO] = open(target, mode="w", encoding="utf-8")
        else:
            dest = sys.stdout
        if lang == "cpp":
            gen = CppCodeGen(
                base,
                dest,
                examples,
                pkg,
                copyright,
                spdx_copyright_text,
                spdx_license_identifier,
            )
            gen.parse(j)
            return
        if lang == "dlang":
            gen = DlangCodeGen(
                base,
                dest,
                examples,
                pkg,
                copyright,
                info,
                salad_version,
            )
            gen.parse(j)
            return
        gen = PythonCodeGen(
            dest, copyright=copyright, parser_info=info, salad_version=salad_version
        )

    elif lang == "java":
        gen = JavaCodeGen(
            base,
            target=target,
            examples=examples,
            package=pkg,
            copyright=copyright,
        )
    elif lang == "typescript":
        gen = TypeScriptCodeGen(base, target=target, package=pkg, examples=examples)
    elif lang == "dotnet":
        gen = DotNetCodeGen(base, target=target, package=pkg, examples=examples)
    else:
        raise SchemaSaladException(f"Unsupported code generation language {lang!r}")

    gen.prologue()

    document_roots = []

    for rec in j:
        if rec["type"] in ("enum", "map", "record", "union"):
            jld = rec.get("jsonldPredicate")
            if isinstance(jld, MutableMapping):
                gen.type_loader(rec, jld.get("_container"), jld.get("noLinkCheck"))
            else:
                gen.type_loader(rec)
            gen.add_vocab(shortname(rec["name"]), rec["name"])

    for rec in j:
        if rec["type"] == "enum":
            for symbol in rec["symbols"]:
                gen.add_vocab(shortname(symbol), symbol)

        if rec["type"] == "record":
            if rec.get("documentRoot"):
                document_roots.append(rec["name"])

            field_names = []
            optional_fields = set()
            for field in rec.get("fields", []):
                field_name = shortname(field["name"])
                field_names.append(field_name)
                tp = field["type"]
                if isinstance(tp, MutableSequence) and tp[0] == "https://w3id.org/cwl/salad#null":
                    optional_fields.add(field_name)

            idfield = ""
            for field in rec.get("fields", []):
                if field.get("jsonldPredicate") == "@id":
                    idfield = field.get("name")

            gen.begin_class(
                rec["name"],
                aslist(rec.get("extends", [])),
                rec.get("doc", ""),
                rec.get("abstract", False),
                field_names,
                idfield,
                optional_fields,
            )
            gen.add_vocab(shortname(rec["name"]), rec["name"])

            sorted_fields = sorted(
                rec.get("fields", []),
                key=lambda i: (
                    FIELD_SORT_ORDER.index(i["name"].split("/")[-1])
                    if i["name"].split("/")[-1] in FIELD_SORT_ORDER
                    else 100
                ),
            )

            for field in sorted_fields:
                if field.get("jsonldPredicate") == "@id":
                    subscope = field.get("subscope")
                    fieldpred = field["name"]
                    optional = bool("https://w3id.org/cwl/salad#null" in field["type"])
                    uri_loader = gen.uri_loader(gen.type_loader(field["type"]), True, False, None)
                    gen.declare_id_field(
                        fieldpred,
                        uri_loader,
                        field.get("doc"),
                        optional,
                    )
                    break

            for field in sorted_fields:
                optional = bool("https://w3id.org/cwl/salad#null" in field["type"])
                jld = field.get("jsonldPredicate")
                fieldpred = field["name"]
                subscope = None

                if isinstance(jld, MutableMapping):
                    type_loader = gen.type_loader(
                        field["type"], jld.get("_container"), jld.get("noLinkCheck")
                    )
                    ref_scope = jld.get("refScope")
                    subscope = jld.get("subscope")
                    if jld.get("typeDSL"):
                        type_loader = gen.typedsl_loader(type_loader, ref_scope)
                    elif jld.get("secondaryFilesDSL"):
                        type_loader = gen.secondaryfilesdsl_loader(type_loader)
                    elif jld.get("_type") == "@id":
                        type_loader = gen.uri_loader(
                            type_loader,
                            jld.get("identity", False),
                            False,
                            ref_scope,
                            jld.get("noLinkCheck"),
                        )
                    elif jld.get("_type") == "@vocab":
                        type_loader = gen.uri_loader(
                            type_loader, False, True, ref_scope, jld.get("noLinkCheck")
                        )

                    map_subject = jld.get("mapSubject")
                    if map_subject:
                        type_loader = gen.idmap_loader(
                            field["name"],
                            type_loader,
                            map_subject,
                            jld.get("mapPredicate"),
                        )

                    if "_id" in jld and jld["_id"][0] != "@":
                        fieldpred = jld["_id"]
                else:
                    type_loader = gen.type_loader(field["type"])

                if jld == "@id":
                    continue

                gen.declare_field(fieldpred, type_loader, field.get("doc"), optional, subscope)

            gen.end_class(rec["name"], field_names)

    root_type = list(document_roots)
    root_type.append({"type": "array", "items": document_roots})

    gen.epilogue(gen.type_loader(root_type))



from typing import Match, Optional

from ..block_parser import BlockParser
from ..core import BlockState
from ..markdown import Markdown
from ._base import BaseDirective, DirectiveParser

__all__ = ["RSTDirective"]

class RSTParser(DirectiveParser):
    name: str
    @staticmethod
    def parse_type(m: Match[str]) -> str: ...
    @staticmethod
    def parse_title(m: Match[str]) -> str: ...
    @staticmethod
    def parse_content(m: Match[str]) -> str: ...

class RSTDirective(BaseDirective):
    parser = RSTParser
    directive_pattern: str
    def parse_directive(
        self, block: BlockParser, m: Match[str], state: BlockState
    ) -> Optional[int]: ...
    def __call__(self, markdown: Markdown) -> None: ...



from typing import Match

from ..block_parser import BlockParser as BlockParser
from ..core import BaseRenderer as BaseRenderer
from ..core import BlockState as BlockState
from ..core import InlineState as InlineState
from ..core import Parser as Parser
from ..helpers import parse_link as parse_link
from ..helpers import parse_link_label as parse_link_label
from ..inline_parser import InlineParser as InlineParser
from ..markdown import Markdown as Markdown
from ..util import unikey as unikey

RUBY_PATTERN: str

def parse_ruby(inline: InlineParser, m: Match[str], state: InlineState) -> int: ...
def render_ruby(renderer: BaseRenderer, text: str, rt: str) -> str: ...
def ruby(md: Markdown) -> None: ...



from typing import Any, Dict, List, Match, Union

from ..block_parser import BlockParser as BlockParser
from ..core import BaseRenderer as BaseRenderer
from ..core import BlockState as BlockState
from ..markdown import Markdown as Markdown
from ._base import BaseDirective as BaseDirective
from ._base import DirectivePlugin as DirectivePlugin

class Include(DirectivePlugin):
    def parse(
        self, block: BlockParser, m: Match[str], state: BlockState
    ) -> Union[Dict[str, Any], List[Dict[str, Any]]]: ...
    def __call__(self, directive: BaseDirective, md: Markdown) -> None: ...

def render_html_include(renderer: BaseRenderer, text: str, **attrs: Any) -> str: ...



from typing import Any, Dict, Match

from _typeshed import Incomplete

from ..block_parser import BlockParser as BlockParser
from ..core import BlockState as BlockState
from ..markdown import Markdown as Markdown
from ._base import BaseDirective as BaseDirective
from ._base import DirectivePlugin as DirectivePlugin

class Admonition(DirectivePlugin):
    SUPPORTED_NAMES: Incomplete
    def parse(self, block: BlockParser, m: Match[str], state: BlockState) -> Dict[str, Any]: ...
    def __call__(self, directive: BaseDirective, md: Markdown) -> None: ...

def render_admonition(self, text: str, name: str, **attrs: Any) -> str: ...
def render_admonition_title(self, text: str) -> str: ...
def render_admonition_content(self, text: str) -> str: ...



from typing import List, Match, Optional

from _typeshed import Incomplete

from ..block_parser import BlockParser
from ..core import BlockState
from ..markdown import Markdown
from ._base import BaseDirective, DirectiveParser, DirectivePlugin

__all__ = ["FencedDirective"]

class FencedParser(DirectiveParser):
    name: str
    @staticmethod
    def parse_type(m: Match[str]) -> str: ...
    @staticmethod
    def parse_title(m: Match[str]) -> str: ...
    @staticmethod
    def parse_content(m: Match[str]) -> str: ...

class FencedDirective(BaseDirective):
    parser = FencedParser
    markers: Incomplete
    directive_pattern: Incomplete
    def __init__(self, plugins: List[DirectivePlugin], markers: str = "`~") -> None: ...
    def parse_directive(
        self, block: BlockParser, m: Match[str], state: BlockState
    ) -> Optional[int]: ...
    def parse_fenced_code(
        self, block: BlockParser, m: Match[str], state: BlockState
    ) -> Optional[int]: ...
    def __call__(self, md: Markdown) -> None: ...



import abc
from abc import ABCMeta, abstractmethod
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Match,
    Optional,
    Tuple,
    Type,
    Union,
)

from ..block_parser import BlockParser as BlockParser
from ..core import BlockState as BlockState
from ..markdown import Markdown as Markdown

class DirectiveParser(ABCMeta, metaclass=abc.ABCMeta):
    name: str
    @staticmethod
    @abstractmethod
    def parse_type(m: Match[str]) -> str: ...
    @staticmethod
    @abstractmethod
    def parse_title(m: Match[str]) -> str: ...
    @staticmethod
    @abstractmethod
    def parse_content(m: Match[str]) -> str: ...
    @classmethod
    def parse_tokens(
        cls, block: BlockParser, text: str, state: BlockState
    ) -> Iterable[Dict[str, Any]]: ...
    @staticmethod
    def parse_options(m: Match[str]) -> List[Tuple[str, str]]: ...

class BaseDirective(metaclass=ABCMeta):
    parser: Type[DirectiveParser]
    directive_pattern: Optional[str]
    def __init__(self, plugins: List["DirectivePlugin"]) -> None: ...
    def register(
        self,
        name: str,
        fn: Callable[
            [BlockParser, Match[str], BlockState], Union[Dict[str, Any], List[Dict[str, Any]]]
        ],
    ) -> None: ...
    def parse_method(
        self, block: BlockParser, m: Match[str], state: BlockState
    ) -> Union[Dict[str, Any], List[Dict[str, Any]]]: ...
    @abstractmethod
    def parse_directive(
        self, block: BlockParser, m: Match[str], state: BlockState
    ) -> Optional[int]: ...
    def register_block_parser(self, md: Markdown, before: Optional[str] = None) -> None: ...
    def __call__(self, markdown: Markdown) -> None: ...

class DirectivePlugin:
    parser: Type[DirectiveParser]
    def __init__(self) -> None: ...
    def parse_options(self, m: Match[str]) -> List[Tuple[str, str]]: ...
    def parse_type(self, m: Match[str]) -> str: ...
    def parse_title(self, m: Match[str]) -> str: ...
    def parse_content(self, m: Match[str]) -> str: ...
    def parse_tokens(
        self, block: BlockParser, text: str, state: BlockState
    ) -> Iterable[Dict[str, Any]]: ...
    def parse(
        self, block: BlockParser, m: Match[str], state: BlockState
    ) -> Union[Dict[str, Any], List[Dict[str, Any]]]: ...
    def __call__(self, directive: BaseDirective, md: Markdown) -> None: ...



from typing import List

from ._base import BaseDirective as BaseDirective
from ._base import DirectiveParser as DirectiveParser
from ._base import DirectivePlugin as DirectivePlugin
from ._fenced import FencedDirective as FencedDirective
from ._rst import RSTDirective as RSTDirective
from .admonition import Admonition as Admonition
from .image import Figure as Figure
from .image import Image as Image
from .include import Include as Include
from .toc import TableOfContents as TableOfContents

__all__ = [
    "DirectiveParser",
    "BaseDirective",
    "DirectivePlugin",
    "RSTDirective",
    "FencedDirective",
    "Admonition",
    "TableOfContents",
    "Include",
    "Image",
    "Figure",
]

class RstDirective(RSTDirective):
    def __init__(self, plugins: List[DirectivePlugin]) -> None: ...



"""
C++17 code generator for a given Schema Salad definition.

Currently only supports emiting YAML from the C++ objects, not yet parsing
YAML into C++ objects.

The generated code requires the libyaml-cpp library & headers

To see an example of usage, look at schema_salad/tests/codegen/cwl.cpp
which can be combined with the CWL V1.0 schema as shown below::

  schema-salad-tool --codegen cpp \
          schema_salad/tests/test_schema/CommonWorkflowLanguage.yml \
          > cwl_v1_0.h

  g++ --std=c++20 -I. -lyaml-cpp schema_salad/tests/codegen/cwl.cpp -o cwl-v1_0-test
  ./cwl-v1_0-test

  # g++ versions older than version 10 may need "--std=c++2a" instead of "--std=c++20"
"""

import os
import re
from typing import IO, Any, Optional, Union, cast

from . import _logger
from .codegen_base import CodeGenBase, TypeDef
from .exceptions import SchemaException
from .schema import shortname
from .utils import aslist


def q(s: str) -> str:
    """Put quotes around a string."""
    return '"' + s + '"'


def replaceKeywords(s: str) -> str:
    """Rename keywords that are reserved in C++."""
    if s in (
        "class",
        "enum",
        "int",
        "long",
        "float",
        "double",
        "default",
        "stdin",
        "stdout",
        "stderr",
        "union",
    ):
        s = s + "_"
    return s


def safename(name: str) -> str:
    """Create a C++ safe name."""
    classname = re.sub("[^a-zA-Z0-9]", "_", name)
    return replaceKeywords(classname)


def safenamespacename(name: str) -> str:
    """Create a C++ safe name for namespaces."""
    name = re.sub("^[a-zA-Z0-9]+://", "", name)  # remove protocol
    name = re.sub("//+", "", name)  # remove duplicate slashes
    name = re.sub("/$", "", name)  # remove trailing slashes
    name = re.sub("[^a-zA-Z0-9/]", "_", name)
    name = re.sub("[/]", "::", name)

    return name


# TODO: this should be somehow not really exists
def safename2(name: dict[str, str]) -> str:
    """Create a namespaced safename."""
    return safenamespacename(name["namespace"]) + "::" + safename(name["classname"])


def split_name(s: str) -> tuple[str, str]:
    """Split url name into its components.

    Splits names like https://xyz.xyz/blub#cwl/class
    into its class path and non class path
    """
    t = s.split("#")
    if len(t) != 2:
        raise ValueError("Expected field to be formatted as 'https://xyz.xyz/blub#cwl/class'.")
    return (t[0], t[1])


def split_field(s: str) -> tuple[str, str, str]:
    """Split field into its components.

    similar to split_name but for field names
    """
    (namespace, field) = split_name(s)
    t = field.split("/")
    if len(t) != 2:
        raise ValueError("Expected field to be formatted as 'https://xyz.xyz/blub#cwl/class'.")
    return (namespace, t[0], t[1])


class ClassDefinition:
    """Prototype of a class."""

    def __init__(self, name: str):
        """Initialize the class definition with a name."""
        self.fullName = name
        self.extends: list[dict[str, str]] = []

        # List of types from parent classes that have been specialized
        self.specializationTypes: list[str] = []

        # this includes fields that are also inheritant
        self.allfields: list[FieldDefinition] = []
        self.fields: list[FieldDefinition] = []
        self.abstract = False
        (self.namespace, self.classname) = split_name(name)
        self.namespace = safenamespacename(self.namespace)
        self.classname = safename(self.classname)

    def writeFwdDeclaration(self, target: IO[str], fullInd: str, ind: str) -> None:
        """Write forward declaration."""
        target.write(f"{fullInd}namespace {self.namespace} {{ struct {self.classname}; }}\n")

    def writeDefinition(
        self, target: IO[Any], fullInd: str, ind: str, common_namespace: str
    ) -> None:
        """Write definition of the class."""
        target.write(f"{fullInd}namespace {self.namespace} {{\n")
        target.write(f"{fullInd}struct {self.classname}")
        extends = list(map(safename2, self.extends))
        override = ""
        virtual = "virtual "
        if len(self.extends) > 0:
            target.write(f"\n{fullInd}{ind}: ")
            target.write(f"\n{fullInd}{ind}, ".join(extends))
            override = " override"
            virtual = ""
        target.write(" {\n")

        for field in self.fields:
            field.writeDefinition(target, fullInd + ind, ind, self.namespace)

        if self.abstract:
            target.write(f"{fullInd}{ind}virtual ~{self.classname}() = 0;\n")
        else:
            target.write(f"{fullInd}{ind}{virtual}~{self.classname}(){override} = default;\n")

        target.write(
            f"{fullInd}{ind}{virtual}auto toYaml([[maybe_unused]] "
            f"{common_namespace}::store_config const& config) const -> YAML::Node{override};\n"
        )
        target.write(f"{fullInd}{ind}{virtual}void fromYaml(YAML::Node const& n){override};\n")
        target.write(f"{fullInd}}};\n")
        target.write(f"{fullInd}}}\n\n")

    def writeImplDefinition(
        self, target: IO[str], fullInd: str, ind: str, common_namespace: str
    ) -> None:
        """Write definition with implementation."""
        extends = list(map(safename2, self.extends))

        # Declaring default destructor
        if self.abstract:
            target.write(
                f"{fullInd}inline {self.namespace}::{self.classname}::~{self.classname}() = default;\n"
            )

        # Write toYaml function
        target.write(
            f"{fullInd}inline auto {self.namespace}::{self.classname}::toYaml([[maybe_unused]] "
            f"::{common_namespace}::store_config const& config) const -> YAML::Node {{\n"
            f"{fullInd}{ind}using ::{common_namespace}::toYaml;\n"
            f"{fullInd}{ind}auto n = YAML::Node{{}};\n"
            f"{fullInd}{ind}if (config.generateTags) {{\n"
            f'{fullInd}{ind}{ind}n.SetTag("{self.classname}");\n'
            f"{fullInd}{ind}}}\n"
        )
        for e in extends:
            target.write(f"{fullInd}{ind}n = mergeYaml(n, {e}::toYaml(config));\n")

        for field in self.fields:
            fieldname = safename(field.name)

            target.write(f"{fullInd}{ind}{{\n")
            target.write(f"{fullInd}{ind}{ind} auto member = toYaml(*{fieldname}, config);\n")
            if field.typeDSL:
                target.write(f"{fullInd}{ind}{ind} member = simplifyType(member, config);\n")
            target.write(
                f"{fullInd}{ind}{ind} member = convertListToMap(member, "
                f"{q(field.mapSubject)}, {q(field.mapPredicate)}, config);\n"
            )
            target.write(f"{fullInd}{ind}{ind}addYamlField(n, {q(field.name)}, member);\n")
            target.write(f"{fullInd}{ind}}}\n")

        target.write(f"{fullInd}{ind}return n;\n{fullInd}}}\n")

        # Write fromYaml function
        functionname = f"{self.namespace}::{self.classname}::fromYaml"
        target.write(
            f"{fullInd}inline void {functionname}([[maybe_unused]] YAML::Node const& n) {{\n"
            f"{fullInd}{ind}using ::{common_namespace}::fromYaml;\n"
        )
        for e in extends:
            target.write(f"{fullInd}{ind}{e}::fromYaml(n);\n")

        for field in self.fields:
            fieldname = safename(field.name)
            expandType = ""
            if field.typeDSL:
                expandType = "expandType"

            target.write(
                f"{fullInd}{ind}{{\n"
                f"{fullInd}{ind}{ind}auto nodeAsList = convertMapToList("
                f"n[{q(field.name)}], {q(field.mapSubject)}, {q(field.mapPredicate)});\n"
                f"{fullInd}{ind}{ind}auto expandedNode = {expandType}(nodeAsList);\n"
                f"{fullInd}{ind}{ind}fromYaml(expandedNode, *{fieldname});\n"
                f"{fullInd}{ind}}}\n"
            )

        target.write(f"{fullInd}}}\n")

        # write type detection function
        if not self.abstract:
            e = f"::{self.namespace}::{self.classname}"
            target.write(
                f"namespace {common_namespace} {{\n"
                f"template <>\n"
                f"struct DetectAndExtractFromYaml<{e}> {{\n"
                f"    auto operator()(YAML::Node const& n) const -> std::optional<{e}> {{\n"
                f"        if (!n.IsDefined()) return std::nullopt;\n"
                f"        if (!n.IsMap()) return std::nullopt;\n"
                f"        auto res = {e}{{}};\n\n"
            )
            for field in self.fields:
                fieldname = safename(field.name)
                target.write(
                    f"        if constexpr (::{common_namespace}::IsConstant<"
                    f"decltype(res.{fieldname})::value_t>::value) try {{\n"
                    f"            fromYaml(n[{q(field.name)}], *res.{fieldname});\n"
                    "            fromYaml(n, res);\n"
                    "            return res;\n"
                    f"        }} catch(...) {{}}\n\n"
                )
            target.write("        return std::nullopt;\n    }\n};\n}\n")


class FieldDefinition:
    """Prototype of a single field from a class definition."""

    def __init__(
        self,
        name: str,
        typeStr: str,
        optional: bool,
        mapSubject: str,
        mapPredicate: str,
        typeDSL: bool,
    ):
        """Initialize field definition.

        Creates a new field with name, its type, optional and which field to use to convert
        from list to map (or empty if it is not possible)
        """
        self.name = name
        self.typeStr = typeStr
        self.optional = optional
        self.mapSubject = mapSubject
        self.mapPredicate = mapPredicate
        self.typeDSL = typeDSL

    def writeDefinition(self, target: IO[Any], fullInd: str, ind: str, namespace: str) -> None:
        """Write a C++ definition for the class field."""
        name = safename(self.name)
        typeStr = self.typeStr.replace(namespace + "::", "")
        target.write(f"{fullInd}heap_object<{typeStr}> {name};\n")


class MapDefinition:
    """Prototype of a map."""

    def __init__(self, name: str, values: list[str]):
        """Initialize union definition with a name and possible values."""
        self.values = values
        (self.namespace, self.classname) = split_name(name)
        self.namespace = safenamespacename(self.namespace)
        self.classname = safename(self.classname)

    def _remove_namespace(self, typeStr: str) -> str:
        return typeStr.replace(f"{self.namespace}::", "")

    def writeFwdDeclaration(self, target: IO[str], fullInd: str, ind: str) -> None:
        """Write forward declaration."""
        target.write(f"{fullInd}namespace {self.namespace} {{ struct {self.classname}; }}\n")

    def writeDefinition(self, target: IO[str], ind: str, common_namespace: str) -> None:
        """Write map definition to output."""
        target.write(f"namespace {self.namespace} {{\n")
        if len(self.values) == 1:
            valueType = self._remove_namespace(self.values[0])
        else:
            valueType = f"std::variant<{', '.join(self._remove_namespace(v) for v in self.values)}>"
        target.write(f"struct {self.classname} {{\n")
        target.write(f"{ind}heap_object<std::map<std::string, {valueType}>> value;\n")
        target.write(
            f"{ind}auto toYaml([[maybe_unused]] "
            f"::{common_namespace}::store_config const& config) const -> YAML::Node;\n"
        )
        target.write(f"{ind}void fromYaml(YAML::Node const& n);\n")
        target.write("};\n")
        target.write("}\n\n")

    def writeImplDefinition(
        self, target: IO[str], fullInd: str, ind: str, common_namespace: str
    ) -> None:
        """Write definition with implementation."""
        # Write toYaml function
        functionname = f"{self.namespace}::{self.classname}::toYaml"
        target.write(
            f"{fullInd}inline auto {functionname}([[maybe_unused]] "
            f"::{common_namespace}::store_config const& config) const -> YAML::Node {{\n"
            f"{fullInd}{ind}using ::{common_namespace}::toYaml;\n"
            f"{fullInd}{ind}return toYaml(*value, config);\n"
            f"{fullInd}}}\n"
        )

        # Write fromYaml function
        functionname = f"{self.namespace}::{self.classname}::fromYaml"
        target.write(
            f"{fullInd}inline void {functionname}([[maybe_unused]] YAML::Node const& n) {{\n"
            f"{fullInd}{ind}using ::{common_namespace}::fromYaml;\n"
            f"{fullInd}{ind}fromYaml(n, *value);\n"
            f"{fullInd}}}\n"
        )


class UnionDefinition:
    """Prototype of a union."""

    def __init__(self, name: str, types: list[str]):
        """Initialize union definition with a name and possible types."""
        (self.namespace, self.classname) = split_name(name)
        self.namespace = safenamespacename(self.namespace)
        self.classname = safename(self.classname)
        self.types = (
            self._remove_namespace(types[0])
            if len(types) == 1
            else f"std::variant<{', '.join(self._remove_namespace(t) for t in types)}>"
        )

    def _remove_namespace(self, typeStr: str) -> str:
        return typeStr.replace(f"{self.namespace}::", "")

    def writeFwdDeclaration(self, target: IO[str], fullInd: str, ind: str) -> None:
        """Write forward declaration."""
        target.write(f"{fullInd}namespace {self.namespace} {{ struct {self.classname}; }}\n")

    def writeDefinition(self, target: IO[str], ind: str, common_namespace: str) -> None:
        """Write union definition to output."""
        target.write(f"namespace {self.namespace} {{\n")
        target.write(f"struct {self.classname} {{\n")
        target.write(f"{ind}{self.types} *value = nullptr;\n")
        target.write(f"{ind}{self.classname}();\n")
        target.write(f"{ind}~{self.classname}();\n")
        target.write(
            f"{ind}auto toYaml([[maybe_unused]] "
            f"::{common_namespace}::store_config const& config) const -> YAML::Node;\n"
        )
        target.write(f"{ind}void fromYaml(YAML::Node const& n);\n")
        target.write("};\n")
        target.write("}\n\n")

    def writeImplDefinition(
        self, target: IO[str], fullInd: str, ind: str, common_namespace: str
    ) -> None:
        """Write definition with implementation."""
        # Write constructor
        functionname = f"{self.namespace}::{self.classname}::{self.classname}"
        target.write(
            f"{fullInd}{functionname}() {{\n"
            f"{fullInd}{ind}value = new {self.types}();\n"
            f"{fullInd}}}\n"
        )

        # Write destructor
        functionname = f"{self.namespace}::{self.classname}::~{self.classname}"
        target.write(
            f"{fullInd}{functionname}() {{\n"
            f"{fullInd}{ind}if (value != nullptr) {{\n"
            f"{fullInd}{ind}{ind}delete value;\n"
            f"{fullInd}{ind}{ind}value = nullptr;\n"
            f"{fullInd}{ind}}}\n"
            f"{fullInd}}}\n"
        )

        # Write toYaml function
        functionname = f"{self.namespace}::{self.classname}::toYaml"
        target.write(
            f"{fullInd}inline auto {functionname}([[maybe_unused]] "
            f"::{common_namespace}::store_config const& config) const -> YAML::Node {{\n"
            f"{fullInd}{ind}using ::{common_namespace}::toYaml;\n"
            f"{fullInd}{ind}return toYaml(*value, config);\n"
            f"{fullInd}}}\n"
        )

        # Write fromYaml function
        functionname = f"{self.namespace}::{self.classname}::fromYaml"
        target.write(
            f"{fullInd}inline void {functionname}([[maybe_unused]] YAML::Node const& n) {{\n"
            f"{fullInd}{ind}using ::{common_namespace}::fromYaml;\n"
            f"{fullInd}{ind}fromYaml(n, *value);\n"
            f"{fullInd}}}\n"
        )


class EnumDefinition:
    """Prototype of a enum."""

    def __init__(self, name: str, values: list[str]):
        """Initialize enum definition with a name and possible values."""
        self.name = name
        self.values = values

        (self.namespace, self.classname) = split_name(name)
        self.namespace = safenamespacename(self.namespace)
        self.classname = safename(self.classname)

    def writeDefinition(self, target: IO[str], ind: str, common_namespace: str) -> None:
        """Write enum definition to output."""
        namespace = ""
        if len(self.name.split("#")) == 2:
            (namespace, classname) = split_name(self.name)
            namespace = safenamespacename(namespace)
            classname = safename(classname)

            name = namespace + "::" + classname
        else:
            name = safename(self.name)
            classname = name
        if len(namespace) > 0:
            target.write(f"namespace {namespace} {{\n")
        target.write(f"enum class {classname} : unsigned int {{\n{ind}")
        target.write(f",\n{ind}".join(map(safename, self.values)))
        target.write("\n};\n")
        target.write(f"inline auto to_string({classname} v) {{\n")
        target.write(f"{ind}static auto m = std::vector<std::string_view> {{\n")
        target.write(f'{ind}    "')
        target.write(f'",\n{ind}    "'.join(self.values))
        target.write(f'"\n{ind}}};\n')

        target.write(f"{ind}using U = std::underlying_type_t<{name}>;\n")
        target.write(f"{ind}return m.at(static_cast<U>(v));\n}}\n")

        if len(namespace) > 0:
            target.write("}\n")

        target.write(f"inline void to_enum(std::string_view v, {name}& out) {{\n")
        target.write(f"{ind}static auto m = std::map<std::string, {name}, std::less<>> {{\n")
        for v in self.values:
            target.write(f"""{ind}{ind}{{{q(v)}, {name}::{safename(v)}}},\n""")
        target.write(f"{ind}}};\n{ind}auto iter = m.find(v);\n")
        target.write(f"{ind}if (iter == m.end()) throw bool{{}};\n")
        target.write(f"{ind}out = iter->second;\n}}\n")

        # Write toYaml function
        target.write(f"namespace {common_namespace} {{\n")
        target.write(
            f"inline auto toYaml({name} v, [[maybe_unused]] "
            f"::{common_namespace}::store_config const& config) {{\n"
        )
        target.write(f"{ind}auto n = YAML::Node{{std::string{{to_string(v)}}}};\n")
        target.write(f'{ind}if (config.generateTags) n.SetTag("{name}");\n')
        target.write(f"{ind}return n;\n}}\n")

        # Write fromYaml function
        target.write(f"inline void fromYaml(YAML::Node n, {name}& out) {{\n")
        target.write(f"{ind}to_enum(n.as<std::string>(), out);\n}}\n")

        if len(self.values):
            target.write(f"template <> struct IsConstant<{name}> : std::true_type {{}};\n")

        target.write("}\n")

        target.write("\n")


# !TODO way to many functions, most of these shouldn't exists
def isPrimitiveType(v: Any) -> bool:
    """Check if v is a primitve type."""
    if not isinstance(v, str):
        return False
    return v in ["null", "boolean", "int", "long", "float", "double", "string"]


def hasFieldValue(e: Any, f: str, v: Any) -> bool:
    """Check if e has a field f value."""
    if not isinstance(e, dict):
        return False
    if f not in e:
        return False
    return bool(e[f] in [v, f"https://w3id.org/cwl/salad#{v}"])


def isRecordSchema(v: Any) -> bool:
    """Check if v is of type record schema."""
    return hasFieldValue(v, "type", "record")


def isEnumSchema(v: Any) -> bool:
    """Check if v is of type enum schema."""
    if not hasFieldValue(v, "type", "enum"):
        return False
    if "symbols" not in v:
        return False
    if not isinstance(v["symbols"], list):
        return False
    return True


def isArray(v: Any) -> bool:
    """Check if v is of type array."""
    if not isinstance(v, list):
        return False
    for i in v:
        if not pred(i):
            return False
    return True


def pred(i: Any) -> bool:
    """Check if v is any of the simple types."""
    return (
        isPrimitiveType(i)
        or isRecordSchema(i)
        or isEnumSchema(i)
        or isArraySchema(i)
        or isMapSchema(i)
        or isUnionSchema(i)
        or isinstance(i, str)
    )


def isArraySchema(v: Any) -> bool:
    """Check if v is of type array schema."""
    if not hasFieldValue(v, "type", "array"):
        return False
    if "items" not in v:
        return False
    if not isinstance(v["items"], list):
        return False

    for i in v["items"]:
        if not (pred(i) or isArray(i)):
            return False
    return True


def isMapSchema(v: Any) -> bool:
    """Check if v is of type map schema."""
    if not hasFieldValue(v, "type", "map"):
        return False
    if "values" not in v:
        return False
    if not isinstance(v["values"], list):
        return False

    for i in v["values"]:
        if not (pred(i) or isArray(i)):
            return False
    return True


def isUnionSchema(v: Any) -> bool:
    """Check if v is of type union schema."""
    return hasFieldValue(v, "type", "union")


class CppCodeGen(CodeGenBase):
    """Generation of C++ code for a given Schema Salad definition."""

    def __init__(
        self,
        base: str,
        target: IO[str],
        examples: Optional[str],
        package: str,
        copyright: Optional[str],
        spdx_copyright_text: Optional[list[str]],
        spdx_license_identifier: Optional[str],
    ) -> None:
        """Initialize the C++ code generator."""
        super().__init__()
        self.base_uri = base
        self.target = target
        self.examples = examples
        self.package = package
        self.copyright = copyright
        self.spdx_copyright_text = spdx_copyright_text
        self.spdx_license_identifier = spdx_license_identifier

        self.classDefinitions: dict[str, ClassDefinition] = {}
        self.enumDefinitions: dict[str, EnumDefinition] = {}
        self.mapDefinitions: dict[str, MapDefinition] = {}
        self.unionDefinitions: dict[str, UnionDefinition] = {}
        self.documentRootTypes: list[ClassDefinition] = []

    def convertTypeToCpp(self, type_declaration: Union[list[Any], dict[str, Any], str]) -> str:
        """Convert a Schema Salad type to a C++ type."""
        if not isinstance(type_declaration, list):
            return self.convertTypeToCpp([type_declaration])

        if len(type_declaration) == 1:
            if type_declaration[0] in ("null", "https://w3id.org/cwl/salad#null"):
                return "std::monostate"
            elif type_declaration[0] in (
                "string",
                "http://www.w3.org/2001/XMLSchema#string",
            ):
                return "std::string"
            elif type_declaration[0] in ("int", "http://www.w3.org/2001/XMLSchema#int"):
                return "int32_t"
            elif type_declaration[0] in (
                "long",
                "http://www.w3.org/2001/XMLSchema#long",
            ):
                return "int64_t"
            elif type_declaration[0] in (
                "float",
                "http://www.w3.org/2001/XMLSchema#float",
            ):
                return "float"
            elif type_declaration[0] in (
                "double",
                "http://www.w3.org/2001/XMLSchema#double",
            ):
                return "double"
            elif type_declaration[0] in (
                "boolean",
                "http://www.w3.org/2001/XMLSchema#boolean",
            ):
                return "bool"
            elif type_declaration[0] == "https://w3id.org/cwl/salad#Any":
                return "std::any"
            elif type_declaration[0] == "https://w3id.org/cwl/cwl#Expression":
                return "cwl_expression_string"
            elif type_declaration[0] in (
                "PrimitiveType",
                "https://w3id.org/cwl/salad#PrimitiveType",
            ):
                return "std::variant<bool, int32_t, int64_t, float, double, std::string>"
            elif isinstance(type_declaration[0], dict):
                if "type" in type_declaration[0] and type_declaration[0]["type"] in (
                    "enum",
                    "https://w3id.org/cwl/salad#enum",
                ):
                    name = type_declaration[0]["name"]
                    if name not in self.enumDefinitions:
                        self.enumDefinitions[name] = EnumDefinition(
                            type_declaration[0]["name"],
                            list(map(shortname, type_declaration[0]["symbols"])),
                        )
                    if len(name.split("#")) != 2:
                        return safename(name)
                    (namespace, classname) = name.split("#")
                    return safenamespacename(namespace) + "::" + safename(classname)
                elif "type" in type_declaration[0] and type_declaration[0]["type"] in (
                    "array",
                    "https://w3id.org/cwl/salad#array",
                ):
                    items = type_declaration[0]["items"]
                    if isinstance(items, list):
                        ts = [self.convertTypeToCpp(i) for i in items]
                        name = ", ".join(ts)
                        return f"std::vector<std::variant<{name}>>"
                    else:
                        i = self.convertTypeToCpp(items)
                        return f"std::vector<{i}>"
                elif "type" in type_declaration[0] and type_declaration[0]["type"] in (
                    "map",
                    "https://w3id.org/cwl/salad#map",
                ):
                    values = type_declaration[0]["values"]
                    if isinstance(values, list):
                        ts = [self.convertTypeToCpp(i) for i in values]
                        name = ", ".join(ts)
                        return f"std::map<std::string, std::variant<{name}>>"
                    else:
                        i = self.convertTypeToCpp(values)
                        return f"std::map<std::string, {i}>"
                elif "type" in type_declaration[0] and type_declaration[0]["type"] in (
                    "record",
                    "https://w3id.org/cwl/salad#record",
                ):
                    n = type_declaration[0]["name"]
                    (namespace, classname) = split_name(n)
                    return safenamespacename(namespace) + "::" + safename(classname)

                n = type_declaration[0]["type"]
                (namespace, classname) = split_name(n)
                return safenamespacename(namespace) + "::" + safename(classname)

            if len(type_declaration[0].split("#")) != 2:
                _logger.debug(f"// something weird2 about {type_declaration[0]}")
                return cast(str, type_declaration[0])

            (namespace, classname) = split_name(type_declaration[0])
            return safenamespacename(namespace) + "::" + safename(classname)

        type_declaration = list(map(self.convertTypeToCpp, type_declaration))
        type_declaration = ", ".join(type_declaration)
        return f"std::variant<{type_declaration}>"

    def epilogue(self, root_loader: Optional[TypeDef]) -> None:
        # find common namespace

        common_namespace = os.path.commonprefix(
            list(
                map(
                    lambda x: x.namespace,
                    list(self.classDefinitions.values()) + list(self.enumDefinitions.values()),
                )
            )
        )
        common_namespace = re.sub("(::)+$", "", common_namespace)

        """Generate final part of our cpp file."""
        if self.spdx_copyright_text:
            for text in self.spdx_copyright_text:
                self.target.write(f"""// SPDX-FileCopyrightText: {text}\n""")

        if self.spdx_license_identifier:
            self.target.write(f"""// SPDX-License-Identifier: {self.spdx_license_identifier}\n""")
        self.target.write("#pragma once\n\n")

        self.target.write(
            """/* This file was generated using schema-salad code generator.
 *
 * The embedded document is subject to the license of the original schema.
 """
        )

        if self.copyright:
            self.target.write("* The original schema is {self.copyright}.\n")

        self.target.write("*/\n\n")

        self.target.write(
            """#include <any>
#include <cassert>
#include <cstddef>
#include <cstdint>
#include <filesystem>
#include <fstream>
#include <map>
#include <optional>
#include <string>
#include <string_view>
#include <variant>
#include <vector>
#include <yaml-cpp/yaml.h>

"""
        )

        self.target.write(f"namespace {common_namespace} {{\n")
        self.target.write(
            """
struct store_config {
    bool simplifyTypes = true;
    bool transformListsToMaps = true;
    bool generateTags = false;
};

inline auto simplifyType(YAML::Node type, store_config const& config) -> YAML::Node {
    if (!config.simplifyTypes) return type;
    auto is_optional = [](YAML::Node const & node) {
        return node.IsSequence() && node.size() == 2u && node[0].Scalar() == "null";
    };

    auto is_array = [](YAML::Node const & node) {
        return node.IsMap() && node["type"].Scalar() == "array" && node["items"].IsScalar();
    };

    // 1. Collapsing optional scalar types into one option
    if (is_optional(type) && type[1].IsScalar()) {
        type = type[1].as<std::string>() + "?";
    }

    // 2. Collapsing array types into one option
    if (is_array(type)) {
        type = type["items"].as<std::string>() + "[]";
    }

    // 3. Collapsing optional array types into one option
    if (is_optional(type) && is_array(type[1])) {
        type = type[1]["items"].as<std::string>() + "[]?";
    }

    return type;
}

inline auto expandType(YAML::Node type) -> YAML::Node {
    auto ends_with = [](std::string str, std::string suffix) {
        if (str.size() < suffix.size()) return false;
        auto str_suffix = str.substr(str.size()-suffix.size(), suffix.size());
        return str_suffix == suffix;
    };

    // 0. If not a scalar type, nothing to do
    if (!type.IsDefined() || !type.IsScalar()) {
        return type;
    }

    auto str = type.as<std::string>();
    // 1. Check if optional array type and expand
    if (ends_with(str, "[]?")) {
        auto result = YAML::Node{};
        result.push_back(YAML::Node{"null"});
        auto array = YAML::Node{};
        array["type"] = "array";
        array["items"] = expandType(YAML::Node(str.substr(0, str.size()-3)));
        result.push_back(array);
        return result;
    }

    // 2. Expand array
    if (ends_with(str, "[]")) {
        auto array = YAML::Node{};
        array["type"] = "array";
        array["items"] = expandType(YAML::Node(str.substr(0, str.size()-2)));
        return array;
    }

    // 3. Expand optional scalar type
    if (ends_with(str, "?")) {
        auto result = YAML::Node{};
        result.push_back(YAML::Node{"null"});
        result.push_back(expandType(YAML::Node(str.substr(0, str.size()-1))));
        return result;
    }
    return type;
}

inline auto mergeYaml(YAML::Node n1, YAML::Node n2) {
    for (auto const& e : n2) {
        n1[e.first.as<std::string>()] = e.second;
    }
    return n1;
}

// declaring toYaml
inline auto toYaml(bool v, [[maybe_unused]] store_config const&) { return YAML::Node{v}; }
inline auto toYaml(float v, [[maybe_unused]] store_config const&) { return YAML::Node{v}; }
inline auto toYaml(double v, [[maybe_unused]] store_config const&) { return YAML::Node{v}; }
inline auto toYaml(char v, [[maybe_unused]] store_config const&) { return YAML::Node{v}; }
inline auto toYaml(int8_t v, [[maybe_unused]] store_config const&) { return YAML::Node{v}; }
inline auto toYaml(uint8_t v, [[maybe_unused]] store_config const&) { return YAML::Node{v}; }
inline auto toYaml(int16_t v, [[maybe_unused]] store_config const&) { return YAML::Node{v}; }
inline auto toYaml(uint16_t v, [[maybe_unused]] store_config const&) { return YAML::Node{v}; }
inline auto toYaml(int32_t v, [[maybe_unused]] store_config const&) { return YAML::Node{v}; }
inline auto toYaml(uint32_t v, [[maybe_unused]] store_config const&) { return YAML::Node{v}; }
inline auto toYaml(int64_t v, [[maybe_unused]] store_config const&) { return YAML::Node{v}; }
inline auto toYaml(uint64_t v, [[maybe_unused]] store_config const&) { return YAML::Node{v}; }
inline auto toYaml(std::monostate const&, [[maybe_unused]] store_config const&) {
    return YAML::Node(YAML::NodeType::Undefined);
}
inline auto toYaml(std::string const& v, [[maybe_unused]] store_config const&) {
    return YAML::Node{v};
}

template <typename T, typename ...Args>
auto anyToYaml_impl(std::any const& a, [[maybe_unused]] store_config const& config) {
    if (auto v = std::any_cast<T const>(&a)) {
        return toYaml(*v, config);
    }
    if constexpr (sizeof...(Args) > 0) {
        return anyToYaml_impl<Args...>(a, config);
    }
    return toYaml(std::monostate{}, config);
}

inline auto toYaml(std::any const& a, [[maybe_unused]] store_config const& config) {
    return anyToYaml_impl<bool,
                          float,
                          double,
                          char,
                          int8_t,
                          uint8_t,
                          int16_t,
                          uint16_t,
                          int32_t,
                          uint32_t,
                          int64_t,
                          uint64_t,
                          std::string>(a, config);
}

// declaring fromYaml
inline void fromYaml(YAML::Node const& n, bool& v) {
    v = n.as<bool>();
}
inline void fromYaml(YAML::Node const& n, float& v) {
    v = n.as<float>();
}
inline void fromYaml(YAML::Node const& n, double& v) {
    v = n.as<double>();
}
inline void fromYaml(YAML::Node const& n, int32_t& v) {
    v = n.as<int32_t>();
}
inline void fromYaml(YAML::Node const& n, int64_t& v) {
    v = n.as<int64_t>();
}
inline void fromYaml(YAML::Node const& n, std::string& v) {
    v = n.as<std::string>();
}
inline void fromYaml(YAML::Node const&, std::any&) {
}
inline void fromYaml(YAML::Node const&, std::monostate&) {
}

inline void addYamlField(YAML::Node& node, std::string const& key, YAML::Node value) {
    if (value.IsDefined()) {
        node[key] = value;
    }
}

inline auto convertListToMap(YAML::Node list, std::string const& mapSubject,
                             std::string const& mapPredicate, store_config const& config) {
    if (!config.transformListsToMaps) return list;
    if (mapSubject.empty()) return list;
    if (list.size() == 0) return list;
    auto map = YAML::Node{};
    for (YAML::Node n : list) {
        auto key = n[mapSubject].as<std::string>();
        if (mapPredicate.empty() || n[mapPredicate].IsMap() || n.size() > 2) {
            n.remove(mapSubject);
            map[key] = n;
        } else {
            map[key] = n[mapPredicate];
        }
    }
    return map;
}
inline auto convertMapToList(YAML::Node map, std::string const& mapSubject,
                             std::string const& mapPredicate) {
    if (mapSubject.empty()) return map;
    if (!map.IsDefined()) return map;
    if (!map.IsMap()) return map;
    auto list = YAML::Node{};
    for (auto n : map) {
        if (mapPredicate.empty() || n.second.IsMap()) {
            n.second[mapSubject] = n.first;
            list.push_back(n.second);
        } else {
            auto n2 = YAML::Node{};
            n2[mapSubject] = n.first;
            n2[mapPredicate] = n.second;
            list.push_back(n2);
        }
    }
    return list;
}

template <typename T> struct IsConstant : std::false_type {};

// fwd declaring toYaml
template <typename T>
auto toYaml(std::vector<T> const& v, [[maybe_unused]] store_config const& config) -> YAML::Node;
template <typename T>
auto toYaml(std::map<std::string, T> const& v, [[maybe_unused]] store_config const& config) -> YAML::Node;
template <typename T>
auto toYaml(T const& t, [[maybe_unused]] store_config const& config) -> YAML::Node;
template <typename ...Args>
auto toYaml(std::variant<Args...> const& t, [[maybe_unused]] store_config const& config) -> YAML::Node;

// fwd declaring fromYaml
template <typename T>
void fromYaml(YAML::Node const& n, std::vector<T>& v);
template <typename T>
void fromYaml(YAML::Node const& n, std::map<std::string, T>& v);
template <typename T>
void fromYaml(YAML::Node const& n, T& t);
template <typename ...Args>
void fromYaml(YAML::Node const& n, std::variant<Args...>& t);

template <typename T>
struct DetectAndExtractFromYaml {
    auto operator()(YAML::Node const&) const -> std::optional<T> {
        return std::nullopt;
    }
};

// special cwl expression string
struct cwl_expression_string {
    std::string s;

    auto toYaml([[maybe_unused]] store_config const& config) const {
        auto n = YAML::Node{s};
        if (config.generateTags) {
            n.SetTag("Expression");
        }
        return n;
    }
    void fromYaml(YAML::Node const& n) {
        s = n.as<std::string>();
    }
};


template <>
struct DetectAndExtractFromYaml<std::monostate> {
    auto operator()(YAML::Node const& n) const -> std::optional<std::monostate> {
        if (!n.IsDefined()) return std::monostate{};
        return std::nullopt;
    }
};

template <typename S>
struct DetectAndExtractFromYaml_implScalar {
    auto operator()(YAML::Node const& n) const -> std::optional<S> {
        try {
            if (n.IsScalar()) return n.as<S>();
        } catch(...) {}
        return std::nullopt;
    }
};

template <> struct DetectAndExtractFromYaml<bool>        : DetectAndExtractFromYaml_implScalar<bool>{};
template <> struct DetectAndExtractFromYaml<float>       : DetectAndExtractFromYaml_implScalar<float>{};
template <> struct DetectAndExtractFromYaml<double>      : DetectAndExtractFromYaml_implScalar<double>{};
template <> struct DetectAndExtractFromYaml<int32_t>     : DetectAndExtractFromYaml_implScalar<int32_t>{};
template <> struct DetectAndExtractFromYaml<int64_t>     : DetectAndExtractFromYaml_implScalar<int64_t>{};
template <> struct DetectAndExtractFromYaml<std::string> : DetectAndExtractFromYaml_implScalar<std::string>{};

template <typename T>
struct DetectAndExtractFromYaml<std::vector<T>> {
    auto operator()(YAML::Node const& n) const -> std::optional<std::vector<T>> {
        if (!n.IsDefined()) return std::nullopt;
        if (!n.IsSequence()) return std::nullopt;
        auto res = std::vector<T>{};
        fromYaml(n, res);
        return res;
    }
};

template <typename T>
struct DetectAndExtractFromYaml<std::map<std::string, T>> {
    auto operator()(YAML::Node const& n) const -> std::optional<std::map<std::string, T>> {
        if (!n.IsDefined()) return std::nullopt;
        if (!n.IsMap()) return std::nullopt;
        auto res = std::map<std::string, T>{};
        fromYaml(n, res);
        return res;
    }
};

template <typename T>
class heap_object {
    std::unique_ptr<T> data = std::make_unique<T>();

public:
    using value_t = T;
    heap_object() noexcept(false) = default;
    heap_object(heap_object const& oth) {
        *data = *oth;
    }
    heap_object(heap_object&& oth) noexcept(noexcept(*data = std::move(*oth))) {
        *data = std::move(*oth);
    }

    template <typename T2>
    heap_object(T2 const& oth) {
        *data = oth;
    }
    template <typename T2>
    heap_object(T2&& oth) noexcept(noexcept(*data = std::forward<T2>(oth))) {
        *data = std::forward<T2>(oth);
    }

    ~heap_object();

    auto operator=(heap_object const& oth) -> heap_object& {
        *data = *oth;
        return *this;
    }
    auto operator=(heap_object&& oth) noexcept(noexcept(*data = std::move(*oth))) -> heap_object& {
        *data = std::move(*oth);
        return *this;
    }

    template <typename T2>
    auto operator=(T2 const& oth) -> heap_object& {
        *data = oth;
        return *this;
    }
    template <typename T2>
    auto operator=(T2&& oth) noexcept(noexcept(*data = std::forward<T2>(oth))) -> heap_object& {
        *data = std::forward<T2>(oth);
        return *this;
    }

    auto operator->() noexcept(true) -> T* {
        return data.get();
    }
    auto operator->() const noexcept(true) -> T const* {
        return data.get();
    }
    auto operator*() noexcept(true) -> T& {
        return *data;
    }
    auto operator*() const noexcept(true) -> T const& {
        return *data;
    }
};

}
"""
        )
        # main body, printing fwd declaration, class definitions, and then implementations

        for key in self.classDefinitions:
            self.classDefinitions[key].writeFwdDeclaration(self.target, "", "    ")
        for key in self.mapDefinitions:
            self.mapDefinitions[key].writeFwdDeclaration(self.target, "", "    ")
        for key in self.unionDefinitions:
            self.unionDefinitions[key].writeFwdDeclaration(self.target, "", "    ")

        # remove parent classes, that are specialized/templated versions
        for key in self.classDefinitions:
            if len(self.classDefinitions[key].specializationTypes) > 0:
                self.classDefinitions[key].extends = []

        # remove fields that are available in a parent class
        for key in self.classDefinitions:
            for field in self.classDefinitions[key].allfields:
                found = False
                for parent_key in self.classDefinitions[key].extends:
                    fullKey = parent_key["namespace"] + "#" + parent_key["classname"]
                    for f in self.classDefinitions[fullKey].allfields:
                        if f.name == field.name:
                            found = True
                            break
                    if found:
                        break

                if not found:
                    self.classDefinitions[key].fields.append(field)  # noqa: B038

        # write definitions
        for key in self.enumDefinitions:
            self.enumDefinitions[key].writeDefinition(self.target, "    ", common_namespace)
        for key in self.classDefinitions:
            self.classDefinitions[key].writeDefinition(self.target, "", "    ", common_namespace)
        for key in self.mapDefinitions:
            self.mapDefinitions[key].writeDefinition(self.target, "    ", common_namespace)
        for key in self.unionDefinitions:
            self.unionDefinitions[key].writeDefinition(self.target, "    ", common_namespace)

        # CPP23: std::unique_ptr in heap_object is constexpr.
        # Hence, the compiler will try to instantiate the destructor on definition.
        # If the destructor was defined inside heap_object, other classes would only
        # be forward declared at this point.
        # This results in an error, because the destructor cannot be generated for
        # incomplete types.
        # Therefore, the destructor is defined here, after all classes have been defined.
        self.target.write(
            f"namespace {common_namespace} {{\n"
            f"template <typename T> heap_object<T>::~heap_object() = default;\n}}\n\n"
        )

        # write implementations
        for key in self.classDefinitions:
            self.classDefinitions[key].writeImplDefinition(
                self.target, "", "    ", common_namespace
            )
        for key in self.mapDefinitions:
            self.mapDefinitions[key].writeImplDefinition(self.target, "", "    ", common_namespace)
        for key in self.unionDefinitions:
            self.unionDefinitions[key].writeImplDefinition(
                self.target, "", "    ", common_namespace
            )

        self.target.write(f"namespace {common_namespace} {{\n")
        self.target.write(
            """
template <typename T>
auto toYaml(std::vector<T> const& v, [[maybe_unused]] store_config const& config) -> YAML::Node {
    auto n = YAML::Node(YAML::NodeType::Sequence);
    for (auto const& e : v) {
        n.push_back(toYaml(e, config));
    }
    return n;
}

template <typename T>
auto toYaml(std::map<std::string, T> const& v, [[maybe_unused]] store_config const& config) -> YAML::Node {
    auto n = YAML::Node(YAML::NodeType::Map);
    for (auto const& [key, value] : v) {
        n[key] = toYaml(value, config);
    }
    return n;
}

template <typename T>
auto toYaml(T const& t, [[maybe_unused]] store_config const& config) -> YAML::Node {
    if constexpr (std::is_enum_v<T>) {
        return toYaml(t, config);
    } else {
        return t.toYaml(config);
    }
}

template <typename ...Args>
auto toYaml(std::variant<Args...> const& t, store_config const& config) -> YAML::Node {
    return std::visit([config](auto const& e) {
        return toYaml(e, config);
    }, t);
}

template <typename T>
void fromYaml(YAML::Node const& n, std::vector<T>& v){
    if (!n.IsSequence()) return;
    for (auto e : n) {
        v.emplace_back();
        fromYaml(e, v.back());
    }
}

template <typename T>
void fromYaml(YAML::Node const& n, std::map<std::string, T>& v){
    if (!n.IsMap()) return;
    for (auto e : n) {
        auto key = e.first.as<std::string>();
        fromYaml(e.second, v[key]);
    }
}

template <typename T>
void fromYaml(YAML::Node const& n, T& t){
    if constexpr (std::is_enum_v<T>) {
        fromYaml(n, t);
    } else {
        t.fromYaml(n);
    }
}

template <typename SomeVariant, typename Head, typename ...Args>
bool detectAndExtractFromYaml(YAML::Node const& n, SomeVariant& v, Head* = nullptr) {
    auto r = DetectAndExtractFromYaml<Head>{}(n);
    if (r) {
        v = *r;
        return true;
    }
    if constexpr (sizeof...(Args) > 0) {
        return detectAndExtractFromYaml<SomeVariant, Args...>(n, v);
    }
    return false;
}

template <typename SomeVariant, typename Head, typename Tail>
bool detectAndExtractFromYaml(YAML::Node const& n, std::variant<std::monostate, Tail>& v, Head* = nullptr) {
    auto r = DetectAndExtractFromYaml<Head>{}(n);
    if (r) {
        v = *r;
        return true;
    }
    auto t = Tail{};
    fromYaml(n, t);
    v = t;
    return true;
}

template <typename ...Args>
void fromYaml(YAML::Node const& n, std::variant<Args...>& v){
    bool found = detectAndExtractFromYaml<std::variant<Args...>, Args...>(n, v);
    if (!found) throw std::runtime_error{"didn't find any overload"};
}
"""
        )
        rootTypes = []
        for cd in self.documentRootTypes:
            rootTypes.append(f"{cd.namespace}::{cd.classname}")
        documentRootType = ", ".join(rootTypes)

        self.target.write(f"using DocumentRootType = std::variant<{documentRootType}>;")
        self.target.write(
            """
auto load_document_from_yaml(YAML::Node n) -> DocumentRootType {
    DocumentRootType root;
    fromYaml(n, root);
    return root;
}
auto load_document_from_string(std::string document) -> DocumentRootType {
    return load_document_from_yaml(YAML::Load(document));
}
auto load_document(std::filesystem::path path) -> DocumentRootType {
    return load_document_from_yaml(YAML::LoadFile(path.string()));
}
void store_document(DocumentRootType const& root, std::ostream& ostream, store_config config={}) {
    auto y = toYaml(root, config);

    YAML::Emitter out;
    out << y;
    ostream << out.c_str() << std::endl;
}
void store_document(DocumentRootType const& root, std::filesystem::path const& path, store_config config={}) {
    auto ofs = std::ofstream{path};
    store_document(root, ofs, config);
}
auto store_document_as_string(DocumentRootType const& root, store_config config={}) -> std::string {
    auto ss = std::stringstream{};
    store_document(root, ss, config);
    return ss.str();
}

}"""
        )

    def parseRecordField(self, field: dict[str, Any]) -> FieldDefinition:
        """Parse a record field."""
        (namespace, classname, fieldname) = split_field(field["name"])
        mapSubject = ""
        mapPredicate = ""
        typeDSL = False
        if "jsonldPredicate" in field:
            if "mapSubject" in field["jsonldPredicate"]:
                mapSubject = field["jsonldPredicate"]["mapSubject"]
            if "mapPredicate" in field["jsonldPredicate"]:
                mapPredicate = field["jsonldPredicate"]["mapPredicate"]
            if "typeDSL" in field["jsonldPredicate"]:
                typeDSL = field["jsonldPredicate"]["typeDSL"]

        if isinstance(field["type"], dict):
            if field["type"]["type"] == "enum":
                fieldtype = "Enum"
            else:
                fieldtype = self.convertTypeToCpp(field["type"])

        else:
            fieldtype = self.convertTypeToCpp(field["type"])

        return FieldDefinition(
            name=fieldname,
            typeStr=fieldtype,
            optional=False,
            mapSubject=mapSubject,
            mapPredicate=mapPredicate,
            typeDSL=typeDSL,
        )

    def parseRecordSchema(self, stype: dict[str, Any]) -> None:
        """Parse a record schema."""
        cd = ClassDefinition(name=stype["name"])
        cd.abstract = stype.get("abstract", False)

        if "extends" in stype:
            for ex in aslist(stype["extends"]):
                (base_namespace, base_classname) = split_name(ex)
                ext = {"namespace": base_namespace, "classname": base_classname}
                cd.extends.append(ext)

        if "specialize" in stype:
            for e in aslist(stype["specialize"]):
                cd.specializationTypes.append(e["specializeFrom"])

        if "fields" in stype:
            for field in stype["fields"]:
                cd.allfields.append(self.parseRecordField(field))

        self.classDefinitions[stype["name"]] = cd

        if stype.get("documentRoot", False):
            self.documentRootTypes.append(cd)

    def parseMapSchema(self, stype: dict[str, Any]) -> str:
        """Parse a map schema."""
        name = cast(str, stype["name"])
        if name not in self.mapDefinitions:
            self.mapDefinitions[name] = MapDefinition(
                name, list(map(self.convertTypeToCpp, stype["values"]))
            )
        return name

    def parseUnionSchema(self, stype: dict[str, Any]) -> str:
        """Parse a union schema."""
        name = cast(str, stype["name"])
        if name not in self.unionDefinitions:
            self.unionDefinitions[name] = UnionDefinition(
                name, list(map(self.convertTypeToCpp, stype["names"]))
            )
        return name

    def parseEnum(self, stype: dict[str, Any]) -> str:
        """Parse a schema salad enum."""
        name = cast(str, stype["name"])
        if name not in self.enumDefinitions:
            self.enumDefinitions[name] = EnumDefinition(
                name, list(map(shortname, stype["symbols"]))
            )
        return name

    def parse(self, items: list[dict[str, Any]]) -> None:
        """Parse sechema salad items.

        This function is being called from the outside and drives
        the whole code generation.
        """
        for stype in items:
            if "type" in stype and stype["type"] == "documentation":
                continue

            if not (pred(stype) or isArray(stype)):
                raise SchemaException("not a valid SaladRecordField")

            # parsing a record
            if isRecordSchema(stype):
                self.parseRecordSchema(stype)
            elif isMapSchema(stype):
                self.parseMapSchema(stype)
            elif isUnionSchema(stype):
                self.parseUnionSchema(stype)
            elif isEnumSchema(stype):
                self.parseEnum(stype)
            else:
                _logger.error(f"not parsed{stype}")

        self.epilogue(None)
        self.target.close()
